from datetime import datetime
from zoneinfo import ZoneInfo

import pytest

from mizani._core.date_utils import (
    align_limits,
    ceil_mid_year,
    ceil_second,
    ceil_week,
    floor_mid_year,
    floor_second,
    floor_week,
)
from mizani._core.dates import (
    datetime_to_num,
    get_tzinfo,
    num_to_datetime,
)


def test_tzinfo():
    tz = ZoneInfo("Africa/Kampala")
    assert get_tzinfo("Africa/Kampala") == tz
    assert get_tzinfo(tz) is tz
    with pytest.raises(TypeError):
        assert get_tzinfo(10)  # type: ignore


def test_floor_mid_year():
    d1 = datetime(2022, 3, 1)
    d2 = datetime(2022, 11, 9)
    assert floor_mid_year(d1) == datetime(2022, 1, 1)
    assert floor_mid_year(d2) == datetime(2022, 7, 1)


def test_ceil_mid_year():
    d1 = datetime(2022, 1, 1)
    d2 = datetime(2022, 1, 2)
    d3 = datetime(2022, 8, 2)
    assert ceil_mid_year(d1) == datetime(2022, 1, 1)
    assert ceil_mid_year(d2) == datetime(2022, 7, 1)
    assert ceil_mid_year(d3) == datetime(2023, 1, 1)


def test_floor_week():
    d1 = datetime(2000, 1, 11)
    d2 = datetime(2000, 8, 21)
    assert floor_week(d1) == datetime(2000, 1, 8)
    assert floor_week(d2) == datetime(2000, 8, 15)


def test_ceil_week():
    d1 = datetime(2000, 1, 15)
    d2 = datetime(2000, 8, 20)
    assert ceil_week(d1) == datetime(2000, 1, 15)
    assert ceil_week(d2) == datetime(2000, 8, 22)


def test_floor_second():
    d1 = datetime(2000, 1, 1, 10, 10, 24, 1000)
    assert floor_second(d1) == datetime(2000, 1, 1, 10, 10, 24)


def test_ceil_second():
    d1 = datetime(2000, 1, 1, 10, 10, 24, 1000)
    assert ceil_second(d1) == datetime(2000, 1, 1, 10, 10, 25)


def test_num_to_datetime():
    limits = num_to_datetime((25552, 27743))
    assert limits[0] == datetime(2039, 12, 17, tzinfo=ZoneInfo("UTC"))
    assert limits[1] == datetime(2045, 12, 16, tzinfo=ZoneInfo("UTC"))

    d = num_to_datetime((27742 + 1.9999999999,))[0]
    assert d.microsecond == 0


def test_datetime_to_num():
    res = datetime_to_num([])
    assert len(res) == 0


# Just for test coverage
# TODO: Find a better test
def test_align_limits():
    limits = (2009, 2010)
    align_limits(limits, 1 + 1e-14)



"""
This module is generated by transpiling Haxe into Python and cleaning
the resulting code by hand, e.g. removing unused Haxe classes. To try it
yourself, clone https://github.com/hsluv/hsluv and run:

    haxe -cp haxe/src hsluv.Hsluv -python hsluv.py
"""
# Last sync version: __version__ = '5.0.3'
# Ref: https://github.com/hsluv/hsluv-python/blob/9c494db/hsluv.py

from __future__ import annotations

import math as _math  # unexport, see #17
from functools import partial as _partial
from functools import wraps as _wraps  # unexport, see #17
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal, TypeAlias

    from mizani.typing import RGBHexColor

    Doublet: TypeAlias = tuple[float, float]
    Triplet: TypeAlias = tuple[float, float, float]
    RGBColor: TypeAlias = tuple[float, float, float]
    Line: TypeAlias = dict[Literal["slope", "intercept"], float]

__all__ = (
    "hex_to_hpluv",
    "hex_to_hsluv",
    "hex_to_rgb",
    "hpluv_to_hex",
    "hpluv_to_lch",
    "hpluv_to_rgb",
    "hsluv_to_hex",
    "hsluv_to_lch",
    "hsluv_to_rgb",
    "lch_to_hpluv",
    "lch_to_hsluv",
    "lch_to_luv",
    "lch_to_rgb",
    "luv_to_lch",
    "luv_to_xyz",
    "rgb_to_hex",
    "rgb_to_hpluv",
    "rgb_to_hsluv",
    "rgb_to_lch",
    "rgb_to_xyz",
    "xyz_to_luv",
    "xyz_to_rgb",
)

# XYZ-to-sRGB matrix
_m = (
    (3.240969941904521, -1.537383177570093, -0.498610760293),
    (-0.96924363628087, 1.87596750150772, 0.041555057407175),
    (0.055630079696993, -0.20397695888897, 1.056971514242878),
)

# sRGB-to-XYZ matrix
_m_inv = (
    (0.41239079926595, 0.35758433938387, 0.18048078840183),
    (0.21263900587151, 0.71516867876775, 0.072192315360733),
    (0.019330818715591, 0.11919477979462, 0.95053215224966),
)
_ref_y = 1.0
_ref_u = 0.19783000664283
_ref_v = 0.46831999493879
_kappa = 903.2962962  # 24389/27 == (29/3)**3
_epsilon = 0.0088564516  # 216/24389 == (6/29)**3


def _normalize_output(conversion):
    # as in snapshot rev 4, the tolerance should be 1e-11
    normalize = _partial(round, ndigits=11 - 1)

    @_wraps(conversion)
    def normalized(*args, **kwargs):
        color = conversion(*args, **kwargs)
        return tuple(normalize(c) for c in color)

    return normalized


def _distance_line_from_origin(line: Line) -> float:
    v = line["slope"] ** 2 + 1
    return abs(line["intercept"]) / _math.sqrt(v)


def _length_of_ray_until_intersect(theta: float, line: Line) -> float:
    return line["intercept"] / (
        _math.sin(theta) - line["slope"] * _math.cos(theta)
    )


def _get_bounds(l: float) -> list[Line]:
    result = []
    sub1 = ((l + 16) ** 3) / 1560896
    sub2 = sub1 if sub1 > _epsilon else l / _kappa

    for m1, m2, m3 in _m:
        for t in (0, 1):
            top1 = (284517 * m1 - 94839 * m3) * sub2
            top2 = (838422 * m3 + 769860 * m2 + 731718 * m1) * l * sub2 - (
                769860 * t
            ) * l
            bottom = (632260 * m3 - 126452 * m2) * sub2 + 126452 * t
            result.append({"slope": top1 / bottom, "intercept": top2 / bottom})
    return result


def _max_safe_chroma_for_l(l: float) -> float:
    return min(_distance_line_from_origin(bound) for bound in _get_bounds(l))


def _max_chroma_for_lh(l: float, h: float) -> float:
    hrad = _math.radians(h)
    lengths = [
        _length_of_ray_until_intersect(hrad, bound) for bound in _get_bounds(l)
    ]
    return min(length for length in lengths if length >= 0)


def _dot_product(a: Triplet, b: Triplet) -> float:
    return sum(i * j for i, j in zip(a, b))


def _from_linear(c: float) -> float:
    if c <= 0.0031308:
        return 12.92 * c

    return 1.055 * (c ** (5 / 12)) - 0.055


def _to_linear(c: float) -> float:
    if c > 0.04045:
        return ((c + 0.055) / 1.055) ** 2.4

    return c / 12.92


def _y_to_l(y: float) -> float:
    if y <= _epsilon:
        return y / _ref_y * _kappa

    return 116 * ((y / _ref_y) ** (1 / 3)) - 16


def _l_to_y(l: float) -> float:
    if l <= 8:
        return _ref_y * l / _kappa

    return _ref_y * (((l + 16) / 116) ** 3)


def xyz_to_rgb(_hx_tuple: Triplet) -> Triplet:
    return (
        _from_linear(_dot_product(_m[0], _hx_tuple)),
        _from_linear(_dot_product(_m[1], _hx_tuple)),
        _from_linear(_dot_product(_m[2], _hx_tuple)),
    )


def rgb_to_xyz(_hx_tuple: Triplet) -> Triplet:
    rgbl = (
        _to_linear(_hx_tuple[0]),
        _to_linear(_hx_tuple[1]),
        _to_linear(_hx_tuple[2]),
    )
    return (
        _dot_product(_m_inv[0], rgbl),
        _dot_product(_m_inv[1], rgbl),
        _dot_product(_m_inv[2], rgbl),
    )


def xyz_to_luv(_hx_tuple: Triplet) -> Triplet:
    x = float(_hx_tuple[0])
    y = float(_hx_tuple[1])
    z = float(_hx_tuple[2])
    l = _y_to_l(y)
    if l == 0:
        return (0, 0, 0)
    divider = x + 15 * y + 3 * z
    if divider == 0:  # pragma: no cover
        u = v = float("nan")
        return (l, u, v)
    var_u = 4 * x / divider
    var_v = 9 * y / divider
    u = 13 * l * (var_u - _ref_u)
    v = 13 * l * (var_v - _ref_v)
    return (l, u, v)


def luv_to_xyz(_hx_tuple: Triplet) -> Triplet:
    l = float(_hx_tuple[0])
    u = float(_hx_tuple[1])
    v = float(_hx_tuple[2])
    if l == 0:
        return (0, 0, 0)
    var_u = u / (13 * l) + _ref_u
    var_v = v / (13 * l) + _ref_v
    y = _l_to_y(l)
    x = y * 9 * var_u / (4 * var_v)
    z = y * (12 - 3 * var_u - 20 * var_v) / (4 * var_v)
    return (x, y, z)


def luv_to_lch(_hx_tuple: Triplet) -> Triplet:
    l = float(_hx_tuple[0])
    u = float(_hx_tuple[1])
    v = float(_hx_tuple[2])
    c = _math.hypot(u, v)
    if c < 1e-08:
        h = 0
    else:
        hrad = _math.atan2(v, u)
        h = _math.degrees(hrad)
        if h < 0:
            h += 360
    return (l, c, h)


def lch_to_luv(_hx_tuple: Triplet) -> Triplet:
    l = float(_hx_tuple[0])
    c = float(_hx_tuple[1])
    h = float(_hx_tuple[2])
    hrad = _math.radians(h)
    u = _math.cos(hrad) * c
    v = _math.sin(hrad) * c
    return (l, u, v)


def hsluv_to_lch(_hx_tuple: Triplet) -> Triplet:
    h = float(_hx_tuple[0])
    s = float(_hx_tuple[1])
    l = float(_hx_tuple[2])
    if l > 100 - 1e-7:
        return (100, 0, h)
    if l < 1e-08:
        return (0, 0, h)
    _hx_max = _max_chroma_for_lh(l, h)
    c = _hx_max / 100 * s
    return (l, c, h)


def lch_to_hsluv(_hx_tuple: Triplet) -> Triplet:
    l = float(_hx_tuple[0])
    c = float(_hx_tuple[1])
    h = float(_hx_tuple[2])
    if l > 100 - 1e-7:
        return (h, 0, 100)
    if l < 1e-08:
        return (h, 0, 0)
    _hx_max = _max_chroma_for_lh(l, h)
    s = c / _hx_max * 100
    return (h, s, l)


def hpluv_to_lch(_hx_tuple: Triplet) -> Triplet:
    h = float(_hx_tuple[0])
    s = float(_hx_tuple[1])
    l = float(_hx_tuple[2])
    if l > 100 - 1e-7:
        return (100, 0, h)
    if l < 1e-08:
        return (0, 0, h)
    _hx_max = _max_safe_chroma_for_l(l)
    c = _hx_max / 100 * s
    return (l, c, h)


def lch_to_hpluv(_hx_tuple: Triplet) -> Triplet:
    l = float(_hx_tuple[0])
    c = float(_hx_tuple[1])
    h = float(_hx_tuple[2])
    if l > 100 - 1e-7:
        return (h, 0, 100)
    if l < 1e-08:
        return (h, 0, 0)
    _hx_max = _max_safe_chroma_for_l(l)
    s = c / _hx_max * 100
    return (h, s, l)


def rgb_to_hex(_hx_tuple: Triplet) -> RGBHexColor:
    return "#{:02x}{:02x}{:02x}".format(
        _math.floor(_hx_tuple[0] * 255 + 0.5),
        _math.floor(_hx_tuple[1] * 255 + 0.5),
        _math.floor(_hx_tuple[2] * 255 + 0.5),
    )


def hex_to_rgb(_hex) -> RGBColor:
    # skip leading '#'
    r = int(_hex[1:3], base=16) / 255.0
    g = int(_hex[3:5], base=16) / 255.0
    b = int(_hex[5:7], base=16) / 255.0
    return (r, g, b)


def lch_to_rgb(_hx_tuple: Triplet) -> RGBColor:
    return xyz_to_rgb(luv_to_xyz(lch_to_luv(_hx_tuple)))


def rgb_to_lch(_hx_tuple: RGBColor) -> Triplet:
    return luv_to_lch(xyz_to_luv(rgb_to_xyz(_hx_tuple)))


def _hsluv_to_rgb(_hx_tuple: Triplet) -> RGBColor:
    return lch_to_rgb(hsluv_to_lch(_hx_tuple))


hsluv_to_rgb = _normalize_output(_hsluv_to_rgb)


def rgb_to_hsluv(_hx_tuple: RGBColor) -> Triplet:
    return lch_to_hsluv(rgb_to_lch(_hx_tuple))


def _hpluv_to_rgb(_hx_tuple: Triplet) -> RGBColor:
    return lch_to_rgb(hpluv_to_lch(_hx_tuple))


hpluv_to_rgb = _normalize_output(_hpluv_to_rgb)


def rgb_to_hpluv(_hx_tuple: RGBColor) -> Triplet:
    return lch_to_hpluv(rgb_to_lch(_hx_tuple))


def hsluv_to_hex(_hx_tuple: Triplet) -> RGBHexColor:
    return rgb_to_hex(hsluv_to_rgb(_hx_tuple))


def hpluv_to_hex(_hx_tuple: Triplet) -> RGBHexColor:
    return rgb_to_hex(hpluv_to_rgb(_hx_tuple))


def hex_to_hsluv(s: RGBHexColor) -> Triplet:
    return rgb_to_hsluv(hex_to_rgb(s))


def hex_to_hpluv(s: RGBHexColor) -> Triplet:
    return rgb_to_hpluv(hex_to_rgb(s))



import numpy as np
import numpy.testing as npt
import pytest

from mizani.palettes import (
    abs_area,
    area_pal,
    brewer_pal,
    cmap_d_pal,
    cmap_pal,
    crayon_palette,
    cubehelix_pal,
    desaturate_pal,
    gradient_n_pal,
    grey_pal,
    hls_palette,
    hsluv_palette,
    hue_pal,
    identity_pal,
    manual_pal,
    rescale_pal,
    xkcd_palette,
)


def test_hls_palette():
    colors = hls_palette(10)
    assert len(colors) == 10
    assert all(len(c) == 3 for c in colors)


def test_hsluv_palette():
    colors = hsluv_palette(5)
    assert len(colors) == 5
    assert all(len(c) == 3 for c in colors)


def test_rescale_pal():
    palette = rescale_pal()
    x = np.arange(0, 1 + 0.01, 0.1)
    result = palette(x)
    assert min(result) == 0.1
    assert max(result) == 1

    palette = rescale_pal((20, 100))
    result = palette(x)
    assert min(result) == 20
    assert max(result) == 100


def test_area_pal():
    palette = area_pal((0, 10))
    x = np.arange(0, 11)
    xsq = (x * 0.1) ** 2
    result = palette(xsq)
    npt.assert_allclose(result, x)


def test_abs_area():
    x = np.arange(0, 1.03, 0.1) ** 2
    palette = abs_area(5)
    result = palette(x)
    assert min(result) == 0
    assert max(result) == 5


def test_grey_pal():
    palette = grey_pal()
    result = palette(5)
    # Same rgb values
    assert all(s[1:3] * 3 == s[1:] for s in result)


def test_hue_pal():
    palette = hue_pal()
    result = palette(5)
    assert all(s[0] == "#" and len(s) == 7 for s in result)

    # branches #
    with pytest.raises(ValueError):
        hue_pal(0.1, 2.3, 3)

    with pytest.raises(ValueError):
        hue_pal(color_space="slh")  # pyright: ignore

    # Backword compatibility check
    palette = hue_pal(color_space="husl")  # pyright: ignore
    result = palette(5)
    assert all(s[0] == "#" and len(s) == 7 for s in result)


def test_brewer_pal():
    result = brewer_pal()(5)
    assert all(s[0] == "#" and len(s) == 7 for s in result)

    result = brewer_pal("qual", 2)(5)
    assert all(s[0] == "#" and len(s) == 7 for s in result)

    result = brewer_pal("div", 2)(5)
    assert all(s[0] == "#" and len(s) == 7 for s in result)

    with pytest.raises(ValueError):
        brewer_pal("div", 200)(5)

    result = brewer_pal("seq", "Greens")(5)
    assert all(s[0] == "#" and len(s) == 7 for s in result)

    with pytest.warns(UserWarning):
        brewer_pal()(100)

    result = brewer_pal("seq", "Blues")(2)
    assert all(s[0] == "#" and len(s) == 7 for s in result)

    result1 = brewer_pal("seq", "Blues", direction=1)(2)
    result2 = brewer_pal("seq", "Blues", direction=-1)(2)
    assert result1 == result2[::-1]

    with pytest.raises(ValueError):
        brewer_pal(direction=-2)(100)


def test_brewer_palette_names():
    from mizani._colors._palettes.brewer import get_palette_names

    names = get_palette_names("sequential")
    assert len(names) > 0

    names = get_palette_names("qualitative")
    assert len(names) > 0

    names = get_palette_names("diverging")
    assert len(names) > 0


def test_brewer_palette_modules():
    from mizani._colors._palettes.brewer import get_palette_module

    with pytest.raises(ValueError):
        get_palette_module("cyclic")


def assert_hex_colors(lst):
    assert all(s[0] == "#" and len(s) == 7 for s in lst)


def test_gradient_n_pal():
    palette = gradient_n_pal(["red", "blue"])
    result = palette([0, 0.25, 0.5, 0.75, 1])
    assert result[0].lower() == "#ff0000"
    assert result[-1].lower() == "#0000ff"
    assert palette([0])[0].lower() == "#ff0000"

    # symmetric gradient
    palette = gradient_n_pal(["red", "blue", "red"], [0, 0.5, 1])
    result = palette([0.2, 0.8])
    assert result[0] == result[1]


def test_cmap_pal():
    palette = cmap_pal("viridis")
    result = palette([0, 0.25, 0.5, 0.75, 1])
    assert_hex_colors(result)

    palette_r = cmap_pal("viridis_r")
    result_r = palette_r([0, 0.25, 0.5, 0.75, 1])
    assert result == result_r[::-1]


def test_cmap_d_pal():
    palette = cmap_d_pal("viridis")
    result = palette(6)
    assert all(s[0] == "#" and len(s) == 7 for s in result)
    assert len(result) == 6

    palette_r = cmap_d_pal("viridis_r")
    result_r = palette_r(6)
    assert result == result_r[::-1]

    # From a row palette
    palette = cmap_d_pal("Accent")
    result = palette(5)
    assert_hex_colors(result)

    # More colors than palette
    result = palette(20)
    assert_hex_colors(result)

    # Bad palette
    with pytest.raises(ValueError):
        palette = cmap_d_pal("Foobar")


def test_desaturate_pal():
    x = [0, 0.25, 0.5, 0.75, 1]
    # When desaturating pure red, green and blue
    # the other 2 colors get matching values,
    # we test that.
    result = desaturate_pal("red", 0.1)(x)
    assert all(s[3:5] == s[5:] for s in result)

    result = desaturate_pal("green", 0.2)(x)
    assert all(s[1:3] == s[-2:] for s in result)

    result = desaturate_pal("blue", 0.3)(x)
    assert all(s[1:3] == s[3:5] for s in result)

    result = desaturate_pal("blue", 0.3, reverse=True)(x)
    assert all(s[1:3] == s[3:5] for s in result)

    with pytest.raises(ValueError):
        desaturate_pal("green", 2.3)


def test_manual_pal():
    size = 5
    n = 3
    values = list(range(size))
    palette = manual_pal(values)
    result = palette(n)
    assert result == values[:n]

    with pytest.warns(UserWarning):
        result = palette(size + 4)


def test_xkcd_palette():
    values = xkcd_palette(["apple green", "red", "tan brown"])
    assert len(values) == 3
    assert_hex_colors(values)


def test_crayon_palette():
    values = crayon_palette(["banana mania", "red", "yellow"])
    assert len(values) == 3
    assert_hex_colors(values)


def test_cubehelix_pal():
    palette = cubehelix_pal()
    values = palette(5)
    assert len(values) == 5
    assert_hex_colors(values)


def test_identity_pal():
    palette = identity_pal()

    x = [1, 2, 3]
    values = palette(x)
    assert values == x

    value = palette(10)
    assert value == 10



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import date, datetime, timedelta, tzinfo
    from types import NoneType
    from typing import (
        Any,
        Callable,
        Literal,
        Optional,
        Protocol,
        Sequence,
        TypeAlias,
        TypedDict,
        TypeVar,
    )

    import numpy as np
    import pandas as pd
    from numpy.typing import NDArray
    from pandas._libs import NaTType

    from mizani._colors._palettes import palette
    from mizani.transforms import trans

    T = TypeVar("T")

    # Tuples
    TupleT2: TypeAlias = tuple[T, T]
    TupleT3: TypeAlias = tuple[T, T, T]
    TupleT4: TypeAlias = tuple[T, T, T, T]
    TupleT5: TypeAlias = tuple[T, T, T, T, T]

    TupleInt2: TypeAlias = TupleT2[int]
    TupleFloat2: TypeAlias = TupleT2[float]
    TupleFloat3: TypeAlias = TupleT3[float]
    TupleFloat4: TypeAlias = TupleT4[float]
    TupleFloat5: TypeAlias = TupleT5[float]
    TupleDatetime2: TypeAlias = TupleT2[datetime]
    TupleDate2: TypeAlias = TupleT2[date]

    # Arrays (strictly numpy)
    NDArrayAny: TypeAlias = NDArray[Any]
    NDArrayBool: TypeAlias = NDArray[np.bool_]
    NDArrayFloat: TypeAlias = NDArray[np.float64]
    NDArrayInt: TypeAlias = NDArray[np.int64]
    NDArrayStr: TypeAlias = NDArray[np.str_]
    NDArrayDatetime: TypeAlias = NDArray[Any]
    NDArrayTimedelta: TypeAlias = NDArray[Any]

    # Series
    AnySeries: TypeAlias = pd.Series[Any]
    BoolSeries: TypeAlias = pd.Series[bool]
    IntSeries: TypeAlias = pd.Series[int]
    FloatSeries: TypeAlias = pd.Series[float]
    DatetimeSeries: TypeAlias = pd.Series[datetime]

    # Use Any as cannot define pd.Series[timedelta]
    TimedeltaSeries: TypeAlias = pd.Series[Any]

    # ArrayLikes
    AnyArrayLike: TypeAlias = NDArrayAny | pd.Series[Any] | Sequence[Any]
    IntArrayLike: TypeAlias = NDArrayInt | IntSeries | Sequence[int]
    FloatArrayLike: TypeAlias = NDArrayFloat | FloatSeries | Sequence[float]
    NumArrayLike: TypeAlias = IntArrayLike | FloatArrayLike
    DatetimeArrayLike: TypeAlias = (
        NDArrayDatetime | DatetimeSeries | Sequence[datetime]
    )
    TimedeltArrayLike: TypeAlias = (
        NDArrayTimedelta | TimedeltaSeries | Sequence[timedelta]
    )

    # Type variable
    TFloatLike = TypeVar("TFloatLike", bound=NDArrayFloat | float)
    TFloatArrayLike = TypeVar("TFloatArrayLike", bound=FloatArrayLike)
    TFloatVector = TypeVar("TFloatVector", bound=NDArrayFloat | FloatSeries)
    TConstrained = TypeVar(
        "TConstrained", int, float, bool, str, complex, datetime, timedelta
    )

    NumericUFunction: TypeAlias = Callable[[TFloatLike], TFloatLike]

    # Nulls for different types
    # float("nan"), np.timedelta64("NaT") & np.datetime64("NaT") do not
    # have distinct types for null
    NullType: TypeAlias = (
        NoneType
        | float
        |
        # Cannot really use NaTType at the moment, e.g. directly
        # instantiating a NaTType is not the same as that from
        # pd.Timestamp("NaT"). Pandas chokes on it.
        NaTType
        | pd.Timestamp
        | pd.Timedelta
        | np.timedelta64
        | np.datetime64
    )

    RGBColor: TypeAlias = tuple[float, float, float] | NDArrayFloat

    RGB256Color: TypeAlias = tuple[int, int, int]
    RGB256Swatch: TypeAlias = Sequence[RGB256Color]
    RGB256Swatches: TypeAlias = Sequence[RGB256Swatch]

    RGBHexColor: TypeAlias = str
    RGBHexSwatch: TypeAlias = Sequence[RGBHexColor]
    RGBHexSwatches: TypeAlias = Sequence[RGBHexSwatch]

    # Change this when numpy gets support for type-hinting shapes
    # Ref: https://github.com/numpy/numpy/issues/16544
    RGBColorArray: TypeAlias = NDArrayFloat

    class SegmentedColorMapData(TypedDict):
        red: Sequence[tuple[float, float, float]]
        green: Sequence[tuple[float, float, float]]
        blue: Sequence[tuple[float, float, float]]

    class SegmentFunctionColorMapData(TypedDict):
        red: Callable[[NDArrayFloat], NDArrayFloat]
        green: Callable[[NDArrayFloat], NDArrayFloat]
        blue: Callable[[NDArrayFloat], NDArrayFloat]

    ColorScheme: TypeAlias = Literal[
        "diverging",
        "qualitative",
        "sequential",
    ]
    ColorSchemeShort: TypeAlias = Literal[
        "div",
        "qual",
        "seq",
    ]
    ColorPalette: TypeAlias = palette

    # Use SI Units where applica
    DurationUnit: TypeAlias = Literal[
        "ns",  # nanosecond
        "us",  # microsecond
        "ms",  # millisecond
        "s",  # second
        "min",  # minute
        "h",  # hour
        "day",  # day
        "week",  # week
        "month",  # month
        "year",  # year
    ]
    Timedelta: TypeAlias = timedelta | pd.Timedelta
    Datetime: TypeAlias = date | datetime | np.datetime64
    SeqDatetime: TypeAlias = (
        Sequence[date] | Sequence[datetime] | Sequence[np.datetime64]
    )
    SeqDatetime64: TypeAlias = Sequence[np.datetime64]
    TzInfo: TypeAlias = tzinfo

    # dateutil.rrule.YEARLY, ..., but not including 2 weekly
    # adding 7 for our own MICROSECONDLY
    DateFreq: TypeAlias = Literal[0, 1, 3, 4, 5, 6, 7]
    DatetimeBreaksUnits: TypeAlias = Literal[
        "auto",
        "microsecond",
        "second",
        "minute",
        "hour",
        "day",
        "week",
        "month",
        "year",
    ]

    ContinuousPalette: TypeAlias = Callable[[FloatArrayLike], Sequence[Any]]
    DiscretePalette: TypeAlias = Callable[[int], NDArrayAny | AnySeries]

    # Mizani
    Trans: TypeAlias = trans
    TransformFunction: TypeAlias = Callable[[AnyArrayLike], NDArrayFloat]
    InverseFunction: TypeAlias = Callable[[FloatArrayLike], NDArrayAny]
    BreaksFunction: TypeAlias = Callable[[tuple[Any, Any]], AnyArrayLike]
    MinorBreaksFunction: TypeAlias = Callable[
        [FloatArrayLike, Optional[TupleFloat2], Optional[int]], NDArrayFloat
    ]

    # Rescale functions
    # This Protocol does not apply to rescale_mid
    class PRescale(Protocol):
        def __call__(
            self,
            x: FloatArrayLike,
            to: TupleFloat2 = (0, 1),
            _from: TupleFloat2 | None = None,
        ) -> NDArrayFloat: ...

    # Censor functions
    class PCensor(Protocol):
        def __call__(
            self,
            x: NDArrayFloat,
            range: TupleFloat2 = (0, 1),
            only_finite: bool = True,
        ) -> NDArrayFloat: ...

    # Any type that has comparison operators can be used to define
    # the domain of a transformation. And implicitly the type of the
    # dataspace.
    class PComparison(Protocol):
        """
        Objects that can be compaired
        """

        def __eq__(self, other, /) -> bool: ...

        def __lt__(self, other, /) -> bool: ...

        def __gt__(self, other, /) -> bool: ...

    DomainType: TypeAlias = TupleT2[PComparison]

    # This does not work probably due to a bug in the typechecker
    # FormatFunction: TypeAlias = Callable[[AnyArrayLike], Sequence[str]]
    FormatFunction: TypeAlias = (
        Callable[[NDArrayAny], Sequence[str]]
        | Callable[[pd.Series[Any]], Sequence[str]]
        | Callable[[Sequence[Any]], Sequence[str]]
    )
    BytesBinarySymbol: TypeAlias = Literal[
        "B",
        "KiB",
        "MiB",
        "GiB",
        "TiB",
        "PiB",
        "EiB",
        "ZiB",
        "YiB",
    ]
    BytesSISymbol: TypeAlias = Literal[
        "B",
        "KB",
        "MB",
        "GB",
        "TB",
        "PB",
        "EB",
        "ZB",
        "YB",
    ]
    BytesSymbol: TypeAlias = BytesBinarySymbol | BytesSISymbol



from __future__ import annotations

from typing import TYPE_CHECKING

from ._colormaps import ColorMap
from ._named_color_values import CRAYON, CSS4, SHORT, XKCD

if TYPE_CHECKING:
    from types import ModuleType

    from mizani.typing import RGBHexColor


__all__ = ("get_colormap", "get_named_color")


class _color_lookup(dict):
    def __getitem__(self, key: str) -> RGBHexColor:
        try:
            return super().__getitem__(key)
        except KeyError as err:
            raise ValueError(f"Unknown name '{key}' for a color.") from err


class _colormap_lookup(dict[str, ColorMap]):
    """
    Lookup (by name) for all available colormaps
    """

    d: dict[str, ColorMap] = {}

    def _lazy_init(self):
        from ._colormaps._maps import (
            _interpolated,
            _listed,
            _palette_interpolated,
            _segment_function,
            _segment_interpolated,
        )

        def _get(mod: ModuleType) -> dict[str, ColorMap]:
            return {name: getattr(mod, name) for name in mod.__all__}

        self.d = {
            **_get(_interpolated),
            **_get(_listed),
            **_get(_palette_interpolated),
            **_get(_segment_function),
            **_get(_segment_interpolated),
        }

    def __getitem__(self, name: str) -> ColorMap:
        if not self.d:
            self._lazy_init()

        try:
            return self.d[name]
        except KeyError as err:
            raise ValueError(f"Unknown colormap: {name}") from err


NAMED_COLORS = _color_lookup(**SHORT, **CSS4, **XKCD, **CRAYON)


def get_named_color(name: str) -> RGBHexColor:
    """
    Return the Hex code of a color
    """
    if name.startswith("#"):
        return name
    else:
        return NAMED_COLORS[name.lower()]


COLORMAPS = _colormap_lookup()


def get_colormap(name: str) -> ColorMap:
    """
    Return colormap
    """
    return COLORMAPS[name]



from datetime import date, datetime, timedelta
from zoneinfo import ZoneInfo

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from mizani._core.dates import _from_ordinalf
from mizani.breaks import (
    breaks_date,
    breaks_extended,
    breaks_log,
    breaks_timedelta,
    minor_breaks,
    minor_breaks_trans,
)
from mizani.transforms import log_trans, trans


def test_log_breaks():
    x = [2, 20, 2000]
    limits = min(x), max(x)
    breaks = breaks_log()(limits)
    npt.assert_array_equal(breaks, [1, 10, 100, 1000, 10000])

    breaks = breaks_log(3)(limits)
    npt.assert_array_equal(breaks, [1, 100, 10000])

    breaks = breaks_log()((10000, 10000))
    npt.assert_array_equal(breaks, [10000])

    breaks = breaks_log()((float("-inf"), float("inf")))
    assert len(breaks) == 0

    # When the limits are in the same order of magnitude
    breaks = breaks_log()([35, 60])
    assert len(breaks) > 0
    assert all(1 < b < 100 for b in breaks)

    breaks = breaks_log()([200, 800])
    npt.assert_array_equal(breaks, [100, 200, 300, 500, 1000])

    breaks = breaks_log()((1664, 14008))
    npt.assert_array_equal(breaks, [1000, 3000, 5000, 10000, 30000])

    breaks = breaks_log()([407, 3430])
    npt.assert_array_equal(breaks, [300, 500, 1000, 3000, 5000])

    breaks = breaks_log()([1761, 8557])
    npt.assert_array_equal(breaks, [1000, 2000, 3000, 5000, 10000])

    # log_breaks -> _log_sub_breaks -> extended_breaks
    breaks = breaks_log(13)([1, 10])
    npt.assert_array_almost_equal(breaks, np.arange(0, 11))

    # No overflow effects
    breaks = breaks_log(n=6)([1e25, 1e30])
    npt.assert_array_almost_equal(breaks, [1e25, 1e26, 1e27, 1e28, 1e29, 1e30])

    # No overflow effects in _log_sub_breaks
    breaks = breaks_log()([2e19, 8e19])
    npt.assert_array_almost_equal(
        breaks, [1.0e19, 2.0e19, 3.0e19, 5.0e19, 1.0e20]
    )

    # _log_sub_breaks for base != 10
    breaks = breaks_log(n=5, base=60)([2e5, 8e5])
    npt.assert_array_almost_equal(
        breaks, [129600, 216000, 432000, 648000, 1080000]
    )

    breaks = breaks_log(n=5, base=2)([20, 80])
    npt.assert_array_almost_equal(breaks, [16, 32, 64, 128])

    # bases & negative breaks
    breaks = breaks_log(base=2)([0.9, 2.9])
    npt.assert_array_almost_equal(breaks, [0.5, 1.0, 2.0, 4.0])


def test_minor_breaks():
    # equidistant breaks
    major = [1, 2, 3, 4]
    limits = [0, 5]
    breaks = minor_breaks()(major, limits)
    npt.assert_array_equal(breaks, [0.5, 1.5, 2.5, 3.5, 4.5])
    minor = minor_breaks(3)(major, [2, 3])
    npt.assert_array_equal(minor, [2.25, 2.5, 2.75])

    # More than 1 minor breaks
    breaks = minor_breaks()(major, limits, 3)
    npt.assert_array_equal(
        breaks,
        [
            0.25,
            0.5,
            0.75,
            1.25,
            1.5,
            1.75,
            2.25,
            2.5,
            2.75,
            3.25,
            3.5,
            3.75,
            4.25,
            4.5,
            4.75,
        ],
    )

    # non-equidistant breaks
    major = [1, 2, 4, 8]
    limits = [0, 10]
    minor = minor_breaks()(major, limits)
    npt.assert_array_equal(minor, [1.5, 3, 6])

    # single major break
    minor = minor_breaks()([2], limits)
    assert len(minor) == 0


def test_minor_breaks_trans():
    major = [1, 2, 3, 4]
    limits = [0, 5]
    regular_mb = minor_breaks()(major, limits)

    class identity_trans(trans):
        transform_is_linear = True
        transform = staticmethod(lambda x: x)
        inverse = staticmethod(lambda x: x)

    t1 = identity_trans()
    t2 = identity_trans()
    t2.minor_breaks = minor_breaks_trans(t2)
    npt.assert_allclose(regular_mb, t1.minor_breaks(major, limits))
    npt.assert_allclose(regular_mb, t2.minor_breaks(major, limits))

    class square_trans(trans):
        transform = staticmethod(np.square)
        inverse = staticmethod(np.sqrt)

    # Transform the input major breaks and check against
    # the inverse of the output minor breaks
    t = square_trans()
    square_mb = t.minor_breaks(np.square(major), np.square(limits))
    npt.assert_allclose(regular_mb, np.sqrt(square_mb))

    # Test minor_breaks for log scales are 2 less than the base
    base = 10
    breaks = np.arange(1, 3)
    limits = [breaks[0], breaks[-1]]
    t = log_trans(base)
    assert len(t.minor_breaks(breaks, limits)) == base - 2

    base = 5  # Odd base
    breaks = np.arange(1, 3)
    limits = [breaks[0], breaks[-1]]
    t = log_trans(base)
    assert len(t.minor_breaks(breaks, limits)) == base - 2

    t = log_trans()
    major = t.transform([1, 10, 100])
    limits = t.transform([1, 100])
    result = minor_breaks_trans(t)(major, limits, n=4)
    npt.assert_allclose(
        result,
        [
            1.02961942,
            1.5260563,
            1.85629799,
            2.10413415,
            3.33220451,
            3.8286414,
            4.15888308,
            4.40671925,
        ],
    )


def test_breaks_date():
    # cpython
    limits = (datetime(2010, 1, 1), datetime(2026, 1, 1))
    breaks = breaks_date("5 Years")
    assert [d.year for d in breaks(limits)] == [2010, 2015, 2020, 2025, 2030]

    breaks = breaks_date("10 Years")(limits)
    assert [d.year for d in breaks] == [2010, 2020, 2030]

    # numpy datetime64
    limits = (np.datetime64("1973"), np.datetime64("1997"))
    breaks = breaks_date(width="10 Years")(limits)
    assert [d.year for d in breaks] == [1970, 1980, 1990, 2000]

    # NaT
    limits = np.datetime64("NaT"), datetime(2017, 1, 1)
    breaks = breaks_date("10 Years")(limits)
    assert len(breaks) == 0

    # automatic monthly breaks
    limits = (datetime(2020, 1, 1), datetime(2021, 1, 15))
    breaks = breaks_date()(limits)
    assert [dt.month for dt in breaks] == [1, 4, 7, 10, 1]

    # automatic monthly breaks with rounding
    limits = (datetime(2019, 12, 27), datetime(2020, 6, 3))
    breaks = breaks_date()(limits)
    assert [dt.month for dt in breaks] == [1, 3, 5]

    # automatic day breaks
    limits = (datetime(2020, 1, 1), datetime(2020, 1, 15))
    breaks = breaks_date()(limits)
    assert [dt.day for dt in breaks] == [1, 5, 9, 13]

    # automatic second breaks
    limits = (datetime(2020, 1, 1, hour=0), datetime(2020, 1, 1, hour=19))
    breaks = breaks_date()(limits)
    assert [dt.hour for dt in breaks] == [0, 4, 8, 12, 16]

    # automatic minute breaks
    limits = (datetime(2020, 1, 1, minute=0), datetime(2020, 1, 1, minute=50))
    breaks = breaks_date()(limits)
    assert [dt.minute for dt in breaks] == [0, 15, 30, 45]

    # automatic second breaks
    limits = (datetime(2020, 1, 1, second=20), datetime(2020, 1, 1, second=50))
    breaks = breaks_date()(limits)
    assert [dt.second for dt in breaks] == [20, 30, 40, 50]

    # automatic microsecond breaks
    limits = (
        datetime(2020, 1, 1, microsecond=10),
        datetime(2020, 1, 1, microsecond=25),
    )
    breaks = breaks_date()(limits)
    assert [dt.microsecond for dt in breaks] == [5, 10, 15, 20, 25, 30]

    # timezone
    UG = ZoneInfo("Africa/Kampala")
    limits = (datetime(1990, 1, 1, tzinfo=UG), datetime(2022, 1, 1, tzinfo=UG))
    breaks = breaks_date()(limits)
    assert breaks[0].tzinfo == UG

    # Special cases
    limits = (datetime(2039, 12, 17), datetime(2045, 12, 16))
    breaks = breaks_date()(limits)
    assert [dt.year for dt in breaks] == [2038, 2040, 2042, 2044, 2046]

    # error cases
    with pytest.raises(ValueError):
        _from_ordinalf(2.933e6, None)


def test_date_type_breaks():
    limits1 = (date(2020, 1, 1), date(2023, 2, 15))
    limits2 = (datetime(2020, 1, 1), datetime(2023, 2, 15))
    calc_breaks = breaks_date()
    res1 = calc_breaks(limits1)
    res2 = [b.replace(tzinfo=None) for b in calc_breaks(limits2)]
    assert res1 == res2


def _check_width(limits, breaks, td: timedelta):
    padding = abs(breaks[0] - limits[0]) + abs(breaks[-1] - limits[-1])
    assert all(np.diff(breaks) == td)
    assert breaks[0] <= limits[0] and limits[-1] <= breaks[-1]
    assert padding < td


def test_breaks_date_week_auto():
    limits = (datetime(2020, 1, 1), datetime(2020, 2, 15))

    breaks = breaks_date()(limits)
    assert {b.day for b in breaks} == {1, 15}

    breaks = breaks_date(n=8)(limits)
    assert {b.day for b in breaks} == {1, 8, 15, 22}


def test_breaks_date_day_width():
    # days
    # 1. The width should be as specified
    # 2. The breaks should encloses the limits
    # 3. The breaks the padding added around the limits should be less than
    #    the width of the breaks
    limits = (datetime(2020, 1, 1), datetime(2020, 2, 28))
    breaks = breaks_date(width="10 days")(limits)
    _check_width(limits, breaks, timedelta(days=10))

    breaks = breaks_date(width="11 days")(limits)
    _check_width(limits, breaks, timedelta(days=11))

    breaks = breaks_date(width="12 days")(limits)
    _check_width(limits, breaks, timedelta(days=12))

    limits = (
        datetime(2000, 1, 1, hour=2),
        datetime(2000, 1, 1, hour=16, second=13),
    )
    breaks = breaks_date(width="2 hour")(limits)
    _check_width(limits, breaks, timedelta(hours=2))

    breaks = breaks_date(width="100 minutes")(limits)
    _check_width(limits, breaks, timedelta(minutes=100))

    breaks = breaks_date(width="5000 seconds")(limits)
    _check_width(limits, breaks, timedelta(seconds=5000))

    breaks = breaks_date(width="5000 seconds")(limits)
    _check_width(limits, breaks, timedelta(seconds=5000))


def test_breaks_date_week_width():
    # weeks
    # 1. The width should be as specified
    # 2. The breaks should encloses the limits
    # 3. The breaks the padding added around the limits should be less than
    #    the width of the breaks
    limits = (datetime(2020, 1, 1), datetime(2020, 2, 28))
    breaks = breaks_date(width="1 week")(limits)
    _check_width(limits, breaks, timedelta(days=7))

    breaks = breaks_date(width="2 weeks")(limits)
    _check_width(limits, breaks, timedelta(days=14))

    breaks = breaks_date(width="3 weeks")(limits)
    _check_width(limits, breaks, timedelta(days=21))


def test_breaks_date_month_width():
    # months
    # 1. The width is within any combination of sequential months
    # 2. The breaks should enclose the limits
    limits = (datetime(2020, 1, 1), datetime(2021, 2, 28))

    breaks = breaks_date(width="1 month")(limits)
    assert all(28 <= d.days <= 31 for d in np.diff(breaks))
    assert breaks[0] <= limits[0] and limits[-1] <= breaks[-1]

    breaks = breaks_date(width="2 month")(limits)
    assert all(59 <= d.days <= 62 for d in np.diff(breaks))
    assert breaks[0] <= limits[0] and limits[-1] <= breaks[-1]
    #
    breaks = breaks_date(width="3 month")(limits)
    assert all(90 <= d.days <= 92 for d in np.diff(breaks))
    assert breaks[0] <= limits[0] and limits[-1] <= breaks[-1]


def test_breaks_timedelta():
    breaks = breaks_timedelta()

    # cpython
    x = [timedelta(days=i * 365) for i in range(25)]
    limits = min(x), max(x)
    major = breaks(limits)
    years = [val.total_seconds() / (365 * 24 * 60 * 60) for val in major]
    npt.assert_array_equal(years, [0, 5, 10, 15, 20, 25])

    x = [timedelta(microseconds=i) for i in range(25)]
    limits = min(x), max(x)
    major = breaks(limits)
    mseconds = [val.total_seconds() * 10**6 for val in major]
    npt.assert_array_equal(mseconds, [0, 5, 10, 15, 20, 25])

    # pandas
    x = [pd.Timedelta(seconds=i * 60) for i in range(10)]
    limits = min(x), max(x)
    major = breaks(limits)
    minutes = [val.total_seconds() / 60 for val in major]
    npt.assert_allclose(minutes, [0, 2, 4, 6, 8])

    # numpy
    x = [np.timedelta64(i * 10, "D") for i in range(1, 10)]
    limits = min(x), max(x)
    with pytest.raises(ValueError):
        breaks(limits)

    # NaT
    limits = pd.NaT, pd.Timedelta(seconds=9 * 60)
    assert len(breaks(limits)) == 0


def test_breaks_extended():
    x = np.arange(100)
    limits = min(x), max(x)
    for n in (5, 7, 10, 13, 31):
        breaks = breaks_extended(n=n)
        assert len(breaks(limits)) <= n + 1

    # Reverse limits
    breaks = breaks_extended(n=7)
    npt.assert_array_equal(breaks((0, 6)), breaks((6, 0)))

    # Infinite limits
    limits = float("-inf"), float("inf")
    breaks = breaks_extended(n=5)
    assert len(breaks(limits)) == 0

    # Zero range discrete
    limits = [1, 1]
    assert len(breaks(limits)) == 1
    assert breaks(limits)[0] == limits[1]

    # Zero range continuous
    limits = [np.pi, np.pi]
    assert len(breaks(limits)) == 1
    assert breaks(limits)[0] == limits[1]



import json
from pathlib import Path

import pytest
from numpy.testing import assert_allclose

from mizani._colors import hsluv

ATOL = 1e-11


def test_rgb_range():
    step = 10
    for h in range(0, 361, step):
        for s in range(0, 101, step):
            for l in range(0, 101, step):
                rgb = hsluv.hsluv_to_rgb((h, s, l))
                for c in rgb:
                    assert -ATOL < c < 1 + ATOL

                rgb = hsluv.hpluv_to_rgb((h, s, l))
                for c in rgb:
                    assert -ATOL < c < 1 + ATOL


@pytest.mark.slow
def test_snapshort():
    # Load snapshot into memory
    filename = Path(__file__).parent / "data/hsluv-snapshot-rev4.json"
    with filename.open() as f:
        snapshot = json.load(f)

    for hex_color, colors in snapshot.items():
        # Test forward functions
        test_rgb = hsluv.hex_to_rgb(hex_color)
        assert_close(test_rgb, colors["rgb"])
        test_xyz = hsluv.rgb_to_xyz(test_rgb)
        assert_close(test_xyz, colors["xyz"])
        test_luv = hsluv.xyz_to_luv(test_xyz)
        assert_close(test_luv, colors["luv"])
        test_lch = hsluv.luv_to_lch(test_luv)
        assert_close(test_lch, colors["lch"])
        test_hsluv = hsluv.lch_to_hsluv(test_lch)
        assert_close(test_hsluv, colors["hsluv"])
        test_hpluv = hsluv.lch_to_hpluv(test_lch)
        assert_close(test_hpluv, colors["hpluv"])

        # Test backward functions
        test_lch = hsluv.hsluv_to_lch(colors["hsluv"])
        assert_close(test_lch, colors["lch"])
        test_lch = hsluv.hpluv_to_lch(colors["hpluv"])
        assert_close(test_lch, colors["lch"])
        test_luv = hsluv.lch_to_luv(test_lch)
        assert_close(test_luv, colors["luv"])
        test_xyz = hsluv.luv_to_xyz(test_luv)
        assert_close(test_xyz, colors["xyz"])
        test_rgb = hsluv.xyz_to_rgb(test_xyz)
        assert_close(test_rgb, colors["rgb"])
        assert hsluv.rgb_to_hex(test_rgb) == hex_color

        # Full test
        assert hsluv.hsluv_to_hex(colors["hsluv"]) == hex_color
        assert_close(hsluv.hex_to_hsluv(hex_color), colors["hsluv"])
        assert hsluv.hpluv_to_hex(colors["hpluv"]) == hex_color
        assert_close(hsluv.hex_to_hpluv(hex_color), colors["hpluv"])


def assert_close(t1, t2):
    assert_allclose(t1, t2, atol=ATOL)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from ..hsluv import hex_to_rgb, rgb_to_hex
from ._colormap import ColorMap, ColorMapKind

if TYPE_CHECKING:
    from typing import Optional, Sequence

    from mizani.typing import (
        FloatArrayLike,
        NDArrayFloat,
        RGBColor,
        RGBColorArray,
        RGBHexColor,
    )


SPACE256 = np.arange(256)
INNER_SPACE256 = SPACE256[1:-1]
ROUNDING_JITTER = 1e-12


class _InterpolatedGen(ColorMap):
    _r_lookup: NDArrayFloat
    _g_lookup: NDArrayFloat
    _b_lookup: NDArrayFloat

    def _generate_colors(self, x: FloatArrayLike) -> Sequence[RGBHexColor]:
        """
        Lookup colors in the interpolated ranges

        Parameters
        ----------
        x :
            Values in the range [0, 1]. O maps to the start of the
            gradient, and 1 to the end of the gradient.
        """
        x = np.asarray(x)
        idx = np.round((x * 255) + ROUNDING_JITTER).astype(int)
        arr = np.column_stack(
            [self._r_lookup[idx], self._g_lookup[idx], self._b_lookup[idx]],
        )
        return [rgb_to_hex(c) for c in arr]


@dataclass
class InterpolatedMap(_InterpolatedGen):
    colors: Sequence[RGBHexColor] | Sequence[RGBColor] | RGBColorArray
    values: Optional[Sequence[float]] = None
    kind: ColorMapKind = ColorMapKind.miscellaneous

    def __post_init__(self):
        from ..named_colors import get_named_color

        if self.values is None:
            values = np.linspace(0, 1, len(self.colors))
        elif len(self.colors) < 2:
            raise ValueError("A color gradient needs two or more colors")
        else:
            values = np.asarray(self.values)
            if values[0] != 0 or values[-1] != 1:
                raise ValueError(
                    "Value points of a color gradient should start"
                    "with 0 and end with 1. "
                    f"Got {values[0]} and {values[-1]}"
                )

        if len(self.colors) != len(values):
            raise ValueError(
                "The values and the colors are different lengths"
                f"colors={len(self.colors)}, values={len(values)}"
            )

        if isinstance(self.colors[0], str):
            colors = [
                hex_to_rgb(get_named_color(c))  # type: ignore
                for c in self.colors
            ]
        else:
            colors = self.colors

        self._data = np.asarray(colors)
        self._r_lookup = interp_lookup(values, self._data[:, 0])
        self._g_lookup = interp_lookup(values, self._data[:, 1])
        self._b_lookup = interp_lookup(values, self._data[:, 2])


def interp_lookup(
    x: NDArrayFloat,
    values: NDArrayFloat,
    values_alt: Optional[NDArrayFloat] = None,
) -> NDArrayFloat:
    """
    Create an interpolation lookup array

    This helps make interpolating between two or more colors
    a discrete task.

    Parameters
    ----------
    x:
        Breaks In the range [0, 1]. Must include 0 and 1 and values
        should be sorted.
    values:
        In the range [0, 1]. Must be the same length as x.
    values_alt:
        In the range [0, 1]. Must be the same length as x.
        Makes it possible to have adjacent interpolation regions
        that with gaps in them numbers. e.g.

            values = [0, 0.1, 0.5, 1]
            values_alt = [0, 0.1, 0.6, 1]

        Creates the regions

            [(0, 0.1), (0.1, 0.5), (0.6, 1)]

        If values_alt is None the region would be

            [(0, 0.1), (0.1, 0.5), (0.5, 1)]
    """
    if values_alt is None:
        values_alt = values

    # - Map x from [0, 1] onto [0, 255] i.e. the color channel
    #   breaks (continuous)
    # - Find where x would be mapped onto the grid (discretizing)
    # - Find the distance between the discrete breaks and the
    #   continuous values of x (with each value scaled by the distance
    #   to previous x value)
    # - Expand the scaled distance (how far to move at each point) to a
    #   value, and move by that scaled distance from the previous point
    x256 = x * 255
    ind = np.searchsorted(x256, SPACE256)[1:-1]
    ind_prev = ind - 1
    distance = (INNER_SPACE256 - x256[ind_prev]) / (x256[ind] - x256[ind_prev])
    stop = values[ind]
    start = values_alt[ind_prev]
    lut = np.concatenate(
        [
            [values[0]],
            start + distance * (stop - start),
            [values[-1]],
        ]
    )
    return np.clip(lut, 0, 1)



"""
Continuous variables have values anywhere in the range minus
infinite to plus infinite. However, when creating a visual
representation of these values what usually matters is the
relative difference between the values. This is where rescaling
comes into play.

The values are mapped onto a range that a scale can deal with. For
graphical representation that range tends to be :math:`[0, 1]` or
:math:`[0, n]`, where :math:`n` is some number that makes the
plotted object overflow the plotting area.

Although a scale may be able handle the :math:`[0, n]` range, it
may be desirable to have a lower bound greater than zero. For
example, if data values get mapped to zero on a scale whose
graphical representation is the size/area/radius/length some data
will be invisible. The solution is to restrict the lower bound
e.g. :math:`[0.1, 1]`. Similarly you can restrict the upper bound
-- using these functions.
"""

from __future__ import annotations

import datetime
import sys
from copy import copy
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from .utils import get_null_value

if TYPE_CHECKING:
    from typing import Any, Optional

    from mizani.typing import (
        FloatArrayLike,
        NDArrayFloat,
        TFloatVector,
        TupleFloat2,
        TupleFloat4,
    )


__all__ = [
    "censor",
    "expand_range",
    "rescale",
    "rescale_max",
    "rescale_mid",
    "squish_infinite",
    "zero_range",
    "expand_range_distinct",
    "squish",
]

EPSILON = sys.float_info.epsilon


def rescale(
    x: FloatArrayLike,
    to: TupleFloat2 = (0, 1),
    _from: Optional[TupleFloat2] = None,
) -> NDArrayFloat:
    """
    Rescale numeric vector to have specified minimum and maximum.

    Parameters
    ----------
    x : array_like | numeric
        1D vector of values to manipulate.
    to : tuple
        output range (numeric vector of length two)
    _from : tuple
        input range (numeric vector of length two).
        If not given, is calculated from the range of x

    Returns
    -------
    out : array_like
        Rescaled values

    Examples
    --------
    >>> x = [0, 2, 4, 6, 8, 10]
    >>> rescale(x)
    array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
    >>> rescale(x, to=(0, 2))
    array([0. , 0.4, 0.8, 1.2, 1.6, 2. ])
    >>> rescale(x, to=(0, 2), _from=(0, 20))
    array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
    """
    __from = (np.min(x), np.max(x)) if _from is None else _from
    return np.interp(x, __from, to)


def rescale_mid(
    x: FloatArrayLike,
    to: TupleFloat2 = (0, 1),
    _from: Optional[TupleFloat2] = None,
    mid: float = 0,
) -> NDArrayFloat:
    """
    Rescale numeric vector to have specified minimum, midpoint,
    and maximum.

    Parameters
    ----------
    x : array_like
        1D vector of values to manipulate.
    to : tuple
        output range (numeric vector of length two)
    _from : tuple
        input range (numeric vector of length two).
        If not given, is calculated from the range of x
    mid	: numeric
        mid-point of input range

    Returns
    -------
    out : array_like
        Rescaled values

    Examples
    --------
    >>> rescale_mid([1, 2, 3], mid=1)
    array([0.5 , 0.75, 1.  ])
    >>> rescale_mid([1, 2, 3], mid=2)
    array([0. , 0.5, 1. ])

    `rescale_mid` does have the same signature as `rescale` and
    `rescale_max`. In cases where we need a compatible function with
    the same signature, we use a closure around the extra `mid` argument.

    >>> def rescale_mid_compat(mid):
    ...     def _rescale(x, to=(0, 1), _from=None):
    ...         return rescale_mid(x, to, _from, mid=mid)
    ...     return _rescale

    >>> rescale_mid2 = rescale_mid_compat(mid=2)
    >>> rescale_mid2([1, 2, 3])
    array([0. , 0.5, 1. ])
    """
    __from: NDArrayFloat = np.array(
        (np.min(x), np.max(x)) if _from is None else _from
    )

    if zero_range(__from) or zero_range(to):  # type: ignore
        out = np.repeat(np.mean(to), len(x))
    else:
        extent = 2 * np.max(np.abs(__from - mid))
        out = (np.asarray(x) - mid) / extent * np.diff(to) + np.mean(to)

    return out


def rescale_max(
    x: FloatArrayLike,
    to: TupleFloat2 = (0, 1),
    _from: Optional[TupleFloat2] = None,
) -> NDArrayFloat:
    """
    Rescale numeric vector to have specified maximum.

    Parameters
    ----------
    x : array_like
        1D vector of values to manipulate.
    to : tuple
        output range (numeric vector of length two)
    _from : tuple
        input range (numeric vector of length two).
        If not given, is calculated from the range of x.
        Only the 2nd (max) element is essential to the
        output.

    Returns
    -------
    out : array_like
        Rescaled values

    Examples
    --------
    >>> x = np.array([0, 2, 4, 6, 8, 10])
    >>> rescale_max(x, (0, 3))
    array([0. , 0.6, 1.2, 1.8, 2.4, 3. ])

    Only the 2nd (max) element of the parameters ``to``
    and ``_from`` are essential to the output.

    >>> rescale_max(x, (1, 3))
    array([0. , 0.6, 1.2, 1.8, 2.4, 3. ])
    >>> rescale_max(x, (0, 20))
    array([ 0.,  4.,  8., 12., 16., 20.])

    If :python:`max(x) < _from[1]` then values will be
    scaled beyond the requested maximum (:python:`to[1]`).

    >>> rescale_max(x, to=(1, 3), _from=(-1, 6))
    array([0., 1., 2., 3., 4., 5.])

    If the values are the same, they taken on the requested maximum.
    This includes an array of all zeros.

    >>> rescale_max(np.array([5, 5, 5]))
    array([1., 1., 1.])
    >>> rescale_max(np.array([0, 0, 0]))
    array([1, 1, 1])
    """
    x = np.asarray(x)
    if _from is None:
        _from = np.min(x), np.max(x)  # type: ignore
        assert _from is not None  # type narrowing

    if np.any(x < 0):
        out = rescale(x, (0, to[1]), _from)
    elif np.all(x == 0) and _from[1] == 0:
        out = np.repeat(to[1], len(x))
    else:
        out = x / _from[1] * to[1]

    return out


def squish_infinite(
    x: FloatArrayLike, range: TupleFloat2 = (0, 1)
) -> NDArrayFloat:
    """
    Truncate infinite values to a range.

    Parameters
    ----------
    x : array_like
        Values that should have infinities squished.
    range : tuple
        The range onto which to squish the infinites.
        Must be of size 2.

    Returns
    -------
    out : array_like
        Values with infinites squished.

    Examples
    --------
    >>> arr1 = np.array([0, .5, .25, np.inf, .44])
    >>> arr2 = np.array([0, -np.inf, .5, .25, np.inf])
    >>> squish_infinite(arr1)
    array([0.  , 0.5 , 0.25, 1.  , 0.44])
    >>> squish_infinite(arr2, (-10, 9))
    array([  0.  , -10.  ,   0.5 ,   0.25,   9.  ])
    """
    _x = np.array(x, copy=True)
    _x[np.isneginf(_x)] = range[0]
    _x[np.isposinf(_x)] = range[1]
    return _x


def squish(
    x: FloatArrayLike, range: TupleFloat2 = (0, 1), only_finite: bool = True
) -> NDArrayFloat:
    """
    Squish values into range.

    Parameters
    ----------
    x : array_like
        Values that should have out of range values squished.
    range : tuple
        The range onto which to squish the values.
    only_finite: boolean
        When true, only squishes finite values.

    Returns
    -------
    out : array_like
        Values with out of range values squished.

    Examples
    --------
    >>> squish([-1.5, 0.2, 0.8, 1.0, 1.2])
    array([0. , 0.2, 0.8, 1. , 1. ])

    >>> squish([-np.inf, -1.5, 0.2, 0.8, 1.0, np.inf], only_finite=False)
    array([0. , 0. , 0.2, 0.8, 1. , 1. ])
    """
    _x = np.array(x, copy=True)
    finite = np.isfinite(_x) if only_finite else True
    _x[np.logical_and(_x < range[0], finite)] = range[0]
    _x[np.logical_and(_x > range[1], finite)] = range[1]
    return _x


def censor(
    x: TFloatVector,
    range: TupleFloat2 = (0, 1),
    only_finite: bool = True,
) -> TFloatVector:
    """
    Convert any values outside of range to a **NULL** type object.

    Parameters
    ----------
    x : array_like
        Values to manipulate
    range : tuple
        (min, max) giving desired output range
    only_finite : bool
        If True (the default), will only modify
        finite values.

    Returns
    -------
    x : array_like
        Censored array

    Examples
    --------
    >>> a = np.array([1, 2, np.inf, 3, 4, -np.inf, 5])
    >>> censor(a, (0, 10))
    array([  1.,   2.,  inf,   3.,   4., -inf,   5.])
    >>> censor(a, (0, 10), False)
    array([ 1.,  2., nan,  3.,  4., nan,  5.])
    >>> censor(a, (2, 4))
    array([ nan,   2.,  inf,   3.,   4., -inf,  nan])

    Notes
    -----
    All values in ``x`` should be of the same type. ``only_finite`` parameter
    is not considered for Datetime and Timedelta types.

    The **NULL** type object depends on the type of values in **x**.

    - :class:`float` - :py:`float('nan')`
    - :class:`int` - :py:`float('nan')`
    - :class:`datetime.datetime` : :py:`np.datetime64(NaT)`
    - :class:`datetime.timedelta` : :py:`np.timedelta64(NaT)`

    """
    res = copy(x)
    if not len(x):
        return res

    null = get_null_value(x)

    if only_finite:
        try:
            finite = np.isfinite(x)
        except TypeError:
            finite = np.repeat(True, len(x))
    else:
        finite = np.repeat(True, len(x))

    # Ignore RuntimeWarning when x contains nans
    with np.errstate(invalid="ignore"):
        outside = (x < range[0]) | (x > range[1])
    bool_idx = finite & outside
    if bool_idx.any():
        if res.dtype == int:
            res = res.astype(float)
        res[bool_idx] = null
    return res


def zero_range(x: tuple[Any, Any], tol: float = EPSILON * 100) -> bool:
    """
    Determine if range of vector is close to zero.

    Parameters
    ----------
    x : array_like
        Value(s) to check. If it is an array_like, it
        should be of length 2.
    tol : float
        Tolerance. Default tolerance is the `machine epsilon`_
        times :math:`10^2`.

    Returns
    -------
    out : bool
        Whether ``x`` has zero range.

    Examples
    --------
    >>> zero_range([1, 1])
    True
    >>> zero_range([1, 2])
    False
    >>> zero_range([1, 2], tol=2)
    True

    .. _machine epsilon: https://en.wikipedia.org/wiki/Machine_epsilon
    """
    if x[0] > x[1]:
        x = x[1], x[0]

    # datetime - pandas, cpython
    if isinstance(x[0], (pd.Timestamp, datetime.datetime)):
        from mizani._core.dates import datetime_to_num

        l, h = datetime_to_num(x)
        return l == h
    # datetime - numpy
    elif isinstance(x[0], np.datetime64):
        return x[0] == x[1]
    # timedelta - pandas, cpython
    elif isinstance(x[0], (pd.Timedelta, datetime.timedelta)):
        return x[0].total_seconds() == x[1].total_seconds()
    # timedelta - numpy
    elif isinstance(x[0], np.timedelta64):
        return x[0] == x[1]
    elif not isinstance(x[0], (float, int, np.number)):
        raise TypeError(
            "zero_range objects cannot work with objects "
            "of type '{}'".format(type(x[0]))
        )
    else:
        low, high = x

    if any(np.isnan((low, high))):
        return True

    if low == high:
        return True

    if any(np.isinf((low, high))):
        return False

    low_abs = np.abs(low)
    if low_abs == 0:
        return False

    return bool(((high - low) / low_abs) < tol)


def expand_range(
    range: TupleFloat2, mul: float = 0, add: float = 0, zero_width: float = 1
) -> TupleFloat2:
    """
    Expand a range with a multiplicative or additive constant

    Parameters
    ----------
    range : tuple
        Range of data. Size 2.
    mul : int | float
        Multiplicative constant
    add : int | float | timedelta
        Additive constant
    zero_width : int | float | timedelta
        Distance to use if range has zero width

    Returns
    -------
    out : tuple
        Expanded range

    Examples
    --------
    >>> expand_range((3, 8))
    (3, 8)
    >>> expand_range((0, 10), mul=0.1)
    (-1.0, 11.0)
    >>> expand_range((0, 10), add=2)
    (-2, 12)
    >>> expand_range((0, 10), mul=.1, add=2)
    (-3.0, 13.0)
    >>> expand_range((0, 1))
    (0, 1)

    When the range has zero width

    >>> expand_range((5, 5))
    (4.5, 5.5)

    Notes
    -----
    If expanding *datetime* or *timedelta* types, **add** and
    **zero_width** must be suitable *timedeltas* i.e. You should
    not mix types between **Numpy**, **Pandas** and the
    :mod:`datetime` module.
    """
    x = range
    low, high = x

    # The expansion cases
    if zero_range(x):
        new = low - zero_width / 2, low + zero_width / 2
    else:
        dx = (high - low) * mul + add
        new = low - dx, high + dx

    return new


def expand_range_distinct(
    range: TupleFloat2,
    expand: TupleFloat2 | TupleFloat4 = (0, 0, 0, 0),
    zero_width: float = 1,
) -> TupleFloat2:
    """
    Expand a range with a multiplicative or additive constants

    Similar to :func:`expand_range` but both sides of the range
    expanded using different constants

    Parameters
    ----------
    range : tuple
        Range of data. Size 2
    expand : tuple
        Length 2 or 4. If length is 2, then the same constants
        are used for both sides. If length is 4 then the first
        two are are the Multiplicative (*mul*) and Additive (*add*)
        constants for the lower limit, and the second two are
        the constants for the upper limit.
    zero_width : int | float | timedelta
        Distance to use if range has zero width

    Returns
    -------
    out : tuple
        Expanded range

    Examples
    --------
    >>> expand_range_distinct((3, 8))
    (3, 8)
    >>> expand_range_distinct((0, 10), (0.1, 0))
    (-1.0, 11.0)
    >>> expand_range_distinct((0, 10), (0.1, 0, 0.1, 0))
    (-1.0, 11.0)
    >>> expand_range_distinct((0, 10), (0.1, 0, 0, 0))
    (-1.0, 10)
    >>> expand_range_distinct((0, 10), (0, 2))
    (-2, 12)
    >>> expand_range_distinct((0, 10), (0, 2, 0, 2))
    (-2, 12)
    >>> expand_range_distinct((0, 10), (0, 0, 0, 2))
    (0, 12)
    >>> expand_range_distinct((0, 10), (.1, 2))
    (-3.0, 13.0)
    >>> expand_range_distinct((0, 10), (.1, 2, .1, 2))
    (-3.0, 13.0)
    >>> expand_range_distinct((0, 10), (0, 0, .1, 2))
    (0, 13.0)
    """

    if len(expand) == 2:
        low_mul = high_mul = expand[0]
        low_add = high_add = expand[1]
    else:
        low_mul, low_add, high_mul, high_add = expand

    lower = expand_range(range, low_mul, low_add, zero_width)[0]
    upper = expand_range(range, high_mul, high_add, zero_width)[1]
    return (lower, upper)



from datetime import date, datetime
from zoneinfo import ZoneInfo

import pandas as pd
import pytest

from mizani.utils import (
    get_categories,
    get_null_value,
    get_timezone,
    match,
    min_max,
    precision,
    round_any,
    same_log10_order_of_magnitude,
)


def test_round_any():
    x = 4.632
    assert round_any(x, 1) == 5
    assert round_any(x, 2) == 4
    assert round_any(x, 3) == 6
    assert round_any(x, 4) == 4
    assert round_any(x, 5) == 5
    assert round_any(x, 1.5) == 4.5


def test_min_max():
    x = [1, 2, 3, 4, 5]
    _min, _max = min_max(x)
    assert _min == 1
    assert _max == 5

    x = [1, float("-inf"), 3, 4, 5]
    _min, _max = min_max(x)
    assert _min == 1
    assert _max == 5

    _min, _max = min_max(x, finite=False)
    assert _min == float("-inf")
    assert _max == 5

    x = [1, 2, float("nan"), 4, 5]
    _min, _max = min_max(x, na_rm=True)
    assert _min == 1
    assert _max == 5

    x = [1, 2, float("nan"), 4, 5, float("inf")]
    _min, _max = min_max(x, na_rm=True, finite=False)
    assert _min == 1
    assert _max == float("inf")

    _min, _max = min_max(x)
    assert str(_min) == "nan"
    assert str(_max) == "nan"

    x = [float("nan"), float("nan"), float("nan")]
    _min, _max = min_max(x, na_rm=True)
    assert _min == float("-inf")
    assert _max == float("inf")


def test_match():
    v1 = [0, 1, 2, 3, 4, 5]
    v2 = [5, 4, 3, 2, 1, 0]
    result = match(v1, v2)
    assert result == v2

    # Positions of the first match
    result = match(v1, v2 + v2)
    assert result == v2

    result = match(v1, v2, incomparables=[1, 2])
    assert result == [5, -1, -1, 2, 1, 0]

    result = match(v1, v2, start=1)
    assert result == [6, 5, 4, 3, 2, 1]

    v2 = [5, 99, 3, 2, 1, 0]
    result = match(v1, v2)
    assert result == [5, 4, 3, 2, -1, 0]


def test_precision():
    assert precision(0.0037) == 1
    assert precision([0.0037, 0.0045]) == 0.0001
    assert precision([0.5, 0.4]) == 0.1
    assert precision([5, 9]) == 1
    assert precision([24, 84]) == 1
    assert precision([290, 784]) == 1
    assert precision([0.0037, 0.5, 9, 24, 784]) == 0.1


def test_same_log10_order_of_magnitude():
    # Default delta
    assert same_log10_order_of_magnitude((2, 8))
    assert same_log10_order_of_magnitude((35, 80.8))
    assert same_log10_order_of_magnitude((232.3, 730))

    assert not same_log10_order_of_magnitude((1, 18))
    assert not same_log10_order_of_magnitude((35, 800))
    assert not same_log10_order_of_magnitude((32, 730))

    assert not same_log10_order_of_magnitude((1, 9.9))
    assert not same_log10_order_of_magnitude((35, 91))
    assert not same_log10_order_of_magnitude((232.3, 950))

    # delta = 0
    assert same_log10_order_of_magnitude((1, 9.9), delta=0)
    assert same_log10_order_of_magnitude((35, 91), delta=0)
    assert same_log10_order_of_magnitude((232.3, 950), delta=0)


def test_get_categories():
    lst = list("abcd")
    s = pd.Series(lst)
    c = pd.Categorical(lst)
    sc = pd.Series(c)

    categories = pd.Index(lst)
    assert categories.equals(get_categories(c))
    assert categories.equals(get_categories(sc))

    with pytest.raises(TypeError):
        assert categories.equals(get_categories(lst))

    with pytest.raises(TypeError):
        assert categories.equals(get_categories(s))


def test_get_timezone():
    UTC = ZoneInfo("UTC")
    UG = ZoneInfo("Africa/Kampala")

    x = [date(2022, 1, 1), date(2022, 12, 1)]
    assert get_timezone(x) is None

    x = [datetime(2022, 1, 1, tzinfo=UTC), datetime(2022, 12, 1, tzinfo=UG)]
    with pytest.warns(UserWarning, match="^Dates"):
        get_timezone(x)


def test_get_null_value():
    x = [datetime(2022, 3, 24)]
    assert get_null_value(x) is None



"""
*"The Grammar of Graphics (2005)"* by **Wilkinson**, **Anand** and
**Grossman** describes three types of transformations.

* *Variable transformations* - Used to make statistical operations on
  variables appropriate and meaningful. They are also used to new
  variables.
* *Scale transformations* - Used to make statistical objects displayed
  on dimensions appropriate and meaningful.
* *Coordinate transformations* - Used to manipulate the geometry of
  graphics to help perceive relationships and find meaningful structures
  for representing variations.

`Variable` and `scale` transformations are similar in-that they lead to
plotted objects that are indistinguishable. Typically, *variable*
transformation is done outside the graphics system and so the system
cannot provide transformation specific guides & decorations for the
plot. The :class:`trans` is aimed at being useful for *scale* and
*coordinate* transformations.
"""

from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from datetime import MAXYEAR, MINYEAR, datetime, timedelta
from types import MethodType
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd

from ._core.dates import datetime_to_num, num_to_datetime
from .breaks import (
    breaks_date,
    breaks_extended,
    breaks_log,
    breaks_symlog,
    breaks_timedelta,
    minor_breaks,
    minor_breaks_trans,
)
from .labels import (
    label_date,
    label_log,
    label_number,
    label_timedelta,
)
from .utils import identity

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, Sequence, Type

    from mizani.typing import (
        BreaksFunction,
        DatetimeArrayLike,
        DomainType,
        FloatArrayLike,
        FormatFunction,
        InverseFunction,
        MinorBreaksFunction,
        NDArrayDatetime,
        NDArrayFloat,
        NDArrayTimedelta,
        TFloatArrayLike,
        TimedeltaSeries,
        TransformFunction,
        TupleFloat2,
    )


__all__ = [
    "asn_trans",
    "atanh_trans",
    "boxcox_trans",
    "modulus_trans",
    "datetime_trans",
    "exp_trans",
    "identity_trans",
    "log10_trans",
    "log1p_trans",
    "log2_trans",
    "log_trans",
    "logit_trans",
    "probability_trans",
    "probit_trans",
    "reverse_trans",
    "sqrt_trans",
    "symlog_trans",
    "timedelta_trans",
    "pd_timedelta_trans",
    "pseudo_log_trans",
    "reciprocal_trans",
    "trans",
    "trans_new",
    "gettrans",
]

UTC = ZoneInfo("UTC")


class trans(ABC):
    """
    Base class for all transforms

    This class is used to transform data and also tell the
    x and y axes how to create and label the tick locations.

    The key methods to override are :meth:`trans.transform`
    and :meth:`trans.inverse`. Alternately, you can quickly
    create a transform class using the :func:`trans_new`
    function.

    Parameters
    ----------
    kwargs : dict
        Attributes of the class to set/override

    """

    #: Whether the transformation over the whole domain is linear.
    #: e.g. `2x` is linear while `1/x` and `log(x)` are not.
    transform_is_linear: bool = False

    domain: DomainType = (-np.inf, np.inf)

    #: Callable to calculate breaks
    breaks_: BreaksFunction = breaks_extended(n=5)

    #: Function to format breaks
    format: FormatFunction = staticmethod(label_number())

    def __init__(self, **kwargs: Any):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise AttributeError(f"Unknown Parameter: {k}")

    @property
    def domain_is_numerical(self) -> bool:
        """
        Return True if transformation acts on numerical data.
        e.g. int, float, and imag are numerical but datetime
        is not.

        """
        return isinstance(self.domain[0], (int, float, np.number))

    def minor_breaks(
        self,
        major: FloatArrayLike,
        limits: Optional[TupleFloat2] = None,
        n: Optional[int] = None,
    ) -> NDArrayFloat:
        """
        Calculate minor_breaks
        """
        n = 1 if n is None else n

        # minor_breaks_trans undoes the transformation and
        # then calculates the breaks. If the domain/dataspace
        # numerical, the calculation will fail.
        if self.transform_is_linear or not self.domain_is_numerical:
            func = minor_breaks(n=n)
        else:
            func = minor_breaks_trans(self, n=n)
        return func(major, limits, n)

    # Use type variables for trans.transform and trans.inverse
    # to help upstream packages avoid type mismatches. e.g.
    # transform(tuple[float, float]) -> tuple[float, float]
    @staticmethod
    @abstractmethod
    def transform(x: TFloatArrayLike) -> TFloatArrayLike:
        """
        Transform of x
        """
        ...

    @staticmethod
    @abstractmethod
    def inverse(x: TFloatArrayLike) -> TFloatArrayLike:
        """
        Inverse of x
        """
        ...

    def breaks(self, limits: DomainType) -> NDArrayFloat:
        """
        Calculate breaks in data space and return them
        in transformed space.

        Expects limits to be in *transform space*, this
        is the same space as that where the domain is
        specified.

        This method wraps around :meth:`breaks_` to ensure
        that the calculated breaks are within the domain
        the transform. This is helpful in cases where an
        aesthetic requests breaks with limits expanded for
        some padding, yet the expansion goes beyond the
        domain of the transform. e.g for a probability
        transform the breaks will be in the domain
        ``[0, 1]`` despite any outward limits.

        Parameters
        ----------
        limits : tuple
            The scale limits. Size 2.

        Returns
        -------
        out : array_like
            Major breaks
        """
        # clip the breaks to the domain,
        # e.g. probabilities will be in [0, 1] domain
        limits = (
            max(self.domain[0], limits[0]),
            min(self.domain[1], limits[1]),
        )
        breaks = np.asarray(self.breaks_(limits))

        # Some methods (e.g. breaks_extended) that
        # calculate breaks take the limits as guide posts and
        # not hard limits.
        breaks = breaks.compress(
            (breaks >= self.domain[0]) & (breaks <= self.domain[1])
        )
        return breaks


def trans_new(
    name: str,
    transform: TransformFunction,
    inverse: InverseFunction,
    breaks: Optional[BreaksFunction] = None,
    minor_breaks: Optional[MinorBreaksFunction] = None,
    _format: Optional[FormatFunction] = None,
    domain=(-np.inf, np.inf),
    doc: str = "",
    **kwargs,
) -> trans:
    """
    Create a transformation class object

    Parameters
    ----------
    name : str
        Name of the transformation
    transform : callable ``f(x)``
        A function (preferably a `ufunc`) that computes
        the transformation.
    inverse : callable ``f(x)``
        A function (preferably a `ufunc`) that computes
        the inverse of the transformation.
    breaks : callable ``f(limits)``
        Function to compute the breaks for this transform.
        If None, then a default good enough for a linear
        domain is used.
    minor_breaks : callable ``f(major, limits)``
        Function to compute the minor breaks for this
        transform. If None, then a default good enough for
        a linear domain is used.
    _format : callable ``f(breaks)``
        Function to format the generated breaks.
    domain : array_like
        Domain over which the transformation is valid.
        It should be of length 2.
    doc : str
        Docstring for the class.
    **kwargs : dict
        Attributes of the transform, e.g if base is passed
        in kwargs, then `t.base` would be a valied attribute.

    Returns
    -------
    out : trans
        Transform class
    """

    def _get(func):
        if isinstance(func, (classmethod, staticmethod, MethodType)):
            return func
        else:
            return staticmethod(func)

    klass_name = "{}_trans".format(name)

    d = {
        "transform": _get(transform),
        "inverse": _get(inverse),
        "domain": domain,
        "__doc__": doc,
        **kwargs,
    }

    if breaks:
        d["breaks_"] = _get(breaks)

    if minor_breaks:
        d["minor_breaks"] = _get(minor_breaks)

    if _format:
        d["format"] = _get(_format)

    return type(klass_name, (trans,), d)  # type: ignore


def log_trans(base: Optional[float] = None, **kwargs: Any) -> trans:
    """
    Create a log transform class for *base*

    Parameters
    ----------
    base : float
        Base for the logarithm. If None, then
        the natural log is used.
    kwargs : dict
        Keyword arguments passed onto
        :func:`trans_new`. Should not include
        the `transform` or `inverse`.

    Returns
    -------
    out : type
        Log transform class
    """
    # transform function
    if base is None:
        name = "log"
        base = np.exp(1)
        transform = np.log  # type: ignore
    elif base == 10:
        name = "log10"
        transform = np.log10  # type: ignore
    elif base == 2:
        name = "log2"
        transform = np.log2  # type: ignore
    else:
        name = "log{}".format(base)

        def transform(x: FloatArrayLike) -> NDArrayFloat:
            return np.log(x) / np.log(base)

    # inverse function
    def inverse(x):
        return np.power(base, x)  # type: ignore

    if "domain" not in kwargs:
        kwargs["domain"] = (sys.float_info.min, np.inf)

    if "breaks" not in kwargs:
        kwargs["breaks"] = breaks_log(base=base)  # type: ignore

    kwargs["base"] = base
    kwargs["_format"] = label_log(base)  # type: ignore

    _trans = trans_new(name, transform, inverse, **kwargs)

    if "minor_breaks" not in kwargs:
        n = int(base) - 2  # type: ignore
        _trans.minor_breaks = minor_breaks_trans(_trans, n=n)

    return _trans


log10_trans = log_trans(10, doc="Log 10 Transformation")
log2_trans = log_trans(2, doc="Log 2 Transformation")


def exp_trans(base: Optional[float] = None, **kwargs: Any):
    """
    Create a exponential transform class for *base*

    This is inverse of the log transform.

    Parameters
    ----------
    base : float
        Base of the logarithm
    kwargs : dict
        Keyword arguments passed onto
        :func:`trans_new`. Should not include
        the `transform` or `inverse`.

    Returns
    -------
    out : type
        Exponential transform class
    """
    # default to e
    if base is None:
        name = "power_e"
        base = np.exp(1)
    else:
        name = "power_{}".format(base)

    # transform function
    def transform(x):
        return np.power(base, x)  # type: ignore

    # inverse function
    def inverse(x):
        return np.log(x) / np.log(base)  # type: ignore

    kwargs["base"] = base
    return trans_new(name, transform, inverse, **kwargs)


class log1p_trans(trans):
    """
    Log plus one Transformation
    """

    transform = staticmethod(np.log1p)  # type: ignore
    inverse = staticmethod(np.expm1)  # type: ignore


class identity_trans(trans):
    """
    Identity Transformation

    Examples
    --------
    The default trans returns one minor break between every pair
    of major break

    >>> major = [0, 1, 2]
    >>> t = identity_trans()
    >>> t.minor_breaks(major)
    array([0.5, 1.5])

    Create a trans that returns 4 minor breaks

    >>> t = identity_trans(minor_breaks=minor_breaks(4))
    >>> t.minor_breaks(major)
    array([0.2, 0.4, 0.6, 0.8, 1.2, 1.4, 1.6, 1.8])
    """

    transform_is_linear = True
    transform = staticmethod(identity)  # type: ignore
    inverse = staticmethod(identity)  # type: ignore


class reverse_trans(trans):
    """
    Reverse Transformation
    """

    transform_is_linear = True
    transform = staticmethod(np.negative)  # type: ignore
    inverse = staticmethod(np.negative)  # type: ignore


class sqrt_trans(trans):
    """
    Square-root Transformation
    """

    transform = staticmethod(np.sqrt)  # type: ignore
    inverse = staticmethod(np.square)  # type: ignore
    domain = (0, np.inf)


class asn_trans(trans):
    """
    Arc-sin square-root Transformation
    """

    @staticmethod
    def transform(x: FloatArrayLike) -> NDArrayFloat:
        return 2 * np.arcsin(np.sqrt(x))  # type: ignore

    @staticmethod
    def inverse(x: FloatArrayLike) -> NDArrayFloat:
        x = np.asarray(x)
        return np.sin(x / 2) ** 2  # type: ignore


class atanh_trans(trans):
    """
    Arc-tangent Transformation
    """

    transform = staticmethod(np.arctanh)  # type: ignore
    inverse = staticmethod(np.tanh)  # type: ignore


def boxcox_trans(p, offset=0, **kwargs):
    r"""
    Boxcox Transformation

    The Box-Cox transformation is a flexible transformation, often
    used to transform data towards normality.

    The Box-Cox power transformation (type 1) requires strictly positive
    values and takes the following form for :math:`y \gt 0`:

    .. math::

        y^{(\lambda)} = \frac{y^\lambda - 1}{\lambda}

    When :math:`y = 0`, the natural log transform is used.

    Parameters
    ----------
    p : float
        Transformation exponent :math:`\lambda`.
    offset : int
        Constant offset. 0 for Box-Cox type 1, otherwise any
        non-negative constant (Box-Cox type 2).
        The default is 0. :func:`~mizani.transforms.modulus_trans`
        sets the default to 1.
    kwargs : dict
        Keyword arguments passed onto :func:`trans_new`. Should not
        include the `transform` or `inverse`.

    References
    ----------
    - Box, G. E., & Cox, D. R. (1964). An analysis of transformations.
      Journal of the Royal Statistical Society. Series B (Methodological),
      211-252. `<https://www.jstor.org/stable/2984418>`_
    - John, J. A., & Draper, N. R. (1980). An alternative family of
      transformations. Applied Statistics, 190-197.
      `<http://www.jstor.org/stable/2986305>`_

    See Also
    --------
    :func:`~mizani.transforms.modulus_trans`

    """

    def transform(x: FloatArrayLike) -> NDArrayFloat:
        x = np.asarray(x)
        if np.any((x + offset) < 0):
            raise ValueError(
                "boxcox_trans must be given only positive values. "
                "Consider using modulus_trans instead?"
            )
        if np.abs(p) < 1e-7:
            return np.log(x + offset)
        else:
            return ((x + offset) ** p - 1) / p

    def inverse(x: FloatArrayLike) -> NDArrayFloat:
        x = np.asarray(x)
        if np.abs(p) < 1e-7:
            return np.exp(x) - offset  # type: ignore
        else:
            return (x * p + 1) ** (1 / p) - offset

    kwargs["p"] = p
    kwargs["offset"] = offset
    kwargs["name"] = kwargs.get("name", "pow_{}".format(p))
    kwargs["transform"] = transform
    kwargs["inverse"] = inverse
    return trans_new(**kwargs)


def modulus_trans(p, offset=1, **kwargs):
    r"""
    Modulus Transformation

    The modulus transformation generalises Box-Cox to work with
    both positive and negative values.

    When :math:`y \neq 0`

    .. math::

        y^{(\lambda)} = sign(y) * \frac{(|y| + 1)^\lambda - 1}{\lambda}

    and when :math:`y = 0`

    .. math::

        y^{(\lambda)} =  sign(y) * \ln{(|y| + 1)}

    Parameters
    ----------
    p : float
        Transformation exponent :math:`\lambda`.
    offset : int
        Constant offset. 0 for Box-Cox type 1, otherwise any
        non-negative constant (Box-Cox type 2).
        The default is 1. :func:`~mizani.transforms.boxcox_trans`
        sets the default to 0.
    kwargs : dict
        Keyword arguments passed onto :func:`trans_new`.
        Should not include the `transform` or `inverse`.

    References
    ----------
    - Box, G. E., & Cox, D. R. (1964). An analysis of transformations.
      Journal of the Royal Statistical Society. Series B (Methodological),
      211-252. `<https://www.jstor.org/stable/2984418>`_
    - John, J. A., & Draper, N. R. (1980). An alternative family of
      transformations. Applied Statistics, 190-197.
      `<http://www.jstor.org/stable/2986305>`_

    See Also
    --------
    :func:`~mizani.transforms.boxcox_trans`
    """
    if np.abs(p) < 1e-7:

        def transform(x: FloatArrayLike) -> NDArrayFloat:
            x = np.asarray(x)
            return np.sign(x) * np.log(np.abs(x) + offset)

        def inverse(x: FloatArrayLike) -> NDArrayFloat:
            x = np.asarray(x)
            return np.sign(x) * (np.exp(np.abs(x)) - offset)  # type: ignore

    else:

        def transform(x: FloatArrayLike) -> NDArrayFloat:
            x = np.asarray(x)
            return np.sign(x) * ((np.abs(x) + offset) ** p - 1) / p

        def inverse(x: FloatArrayLike) -> NDArrayFloat:
            x = np.asarray(x)
            return np.sign(x) * ((np.abs(x) * p + 1) ** (1 / p) - offset)

    kwargs["p"] = p
    kwargs["offset"] = offset
    kwargs["name"] = kwargs.get("name", "mt_pow_{}".format(p))
    kwargs["transform"] = transform
    kwargs["inverse"] = inverse
    return trans_new(**kwargs)


def probability_trans(distribution: str, *args, **kwargs) -> trans:
    """
    Probability Transformation

    Parameters
    ----------
    distribution : str
        Name of the distribution. Valid distributions are
        listed at :mod:`scipy.stats`. Any of the continuous
        or discrete distributions.
    args : tuple
        Arguments passed to the distribution functions.
    kwargs : dict
        Keyword arguments passed to the distribution functions.

    Notes
    -----
    Make sure that the distribution is a good enough
    approximation for the data. When this is not the case,
    computations may run into errors. Absence of any errors
    does not imply that the distribution fits the data.
    """
    import scipy.stats as stats

    cdists = {k for k in dir(stats) if hasattr(getattr(stats, k), "cdf")}
    if distribution not in cdists:
        raise ValueError(f"Unknown distribution '{distribution}'")

    try:
        doc = kwargs.pop("_doc")
    except KeyError:
        doc = ""

    try:
        name = kwargs.pop("_name")
    except KeyError:
        name = "prob_{}".format(distribution)

    def transform(x: FloatArrayLike) -> NDArrayFloat:
        return getattr(stats, distribution).cdf(x, *args, **kwargs)

    def inverse(x: FloatArrayLike) -> NDArrayFloat:
        return getattr(stats, distribution).ppf(x, *args, **kwargs)

    return trans_new(name, transform, inverse, domain=(0, 1), doc=doc)


logit_trans = probability_trans(
    "logistic", _name="logit", _doc="Logit Transformation"
)
probit_trans = probability_trans(
    "norm", _name="norm", _doc="Probit Transformation"
)


class datetime_trans(trans):
    """
    Datetime Transformation

    Parameters
    ----------
    tz : str | ZoneInfo
        Timezone information

    Examples
    --------
    >>> from zoneinfo import ZoneInfo
    >>> UTC = ZoneInfo("UTC")
    >>> EST = ZoneInfo("EST")
    >>> t = datetime_trans(EST)
    >>> x = [datetime(2022, 1, 20, tzinfo=UTC)]
    >>> x2 = t.inverse(t.transform(x))
    >>> list(x) == list(x2)
    True
    >>> x[0].tzinfo == x2[0].tzinfo
    False
    >>> x[0].tzinfo.key
    'UTC'
    >>> x2[0].tzinfo.key
    'EST'
    """

    domain = (
        datetime(MINYEAR, 1, 1, tzinfo=UTC),
        datetime(MAXYEAR, 12, 31, tzinfo=UTC),
    )
    breaks_ = staticmethod(breaks_date())
    format = staticmethod(label_date())
    tz = None

    def __init__(self, tz=None, **kwargs):
        if isinstance(tz, str):
            tz = ZoneInfo(tz)

        super().__init__(**kwargs)
        self.tz = tz

    def transform(self, x: DatetimeArrayLike) -> NDArrayFloat:  # pyright: ignore[reportIncompatibleMethodOverride]
        """
        Transform from date to a numerical format
        """
        if not len(x):
            return np.array([])

        x0 = next(iter(x))
        try:
            tz = x0.tzinfo
        except AttributeError:
            tz = None

        if tz and self.tz is None:
            self.tz = tz

        return datetime_to_num(x)  # type: ignore

    def inverse(self, x: FloatArrayLike) -> NDArrayDatetime:  # pyright: ignore[reportIncompatibleMethodOverride]
        """
        Transform to date from numerical format
        """
        return num_to_datetime(x, tz=self.tz)

    @property
    def tzinfo(self):
        """
        Alias of `tz`
        """
        return self.tz


class timedelta_trans(trans):
    """
    Timedelta Transformation
    """

    domain = (timedelta.min, timedelta.max)
    breaks_ = staticmethod(breaks_timedelta())
    format = staticmethod(label_timedelta())

    @staticmethod
    def transform(x: NDArrayTimedelta | Sequence[timedelta]) -> NDArrayFloat:  # pyright: ignore[reportIncompatibleMethodOverride]
        """
        Transform from Timeddelta to numerical format
        """
        # microseconds
        return np.array([_x.total_seconds() * 10**6 for _x in x])

    @staticmethod
    def inverse(x: FloatArrayLike) -> NDArrayTimedelta:
        """
        Transform to Timedelta from numerical format
        """
        return np.array([timedelta(microseconds=i) for i in x])


class pd_timedelta_trans(trans):
    """
    Pandas timedelta Transformation
    """

    domain = (pd.Timedelta.min, pd.Timedelta.max)
    breaks_ = staticmethod(breaks_timedelta())
    format = staticmethod(label_timedelta())

    @staticmethod
    def transform(x: TimedeltaSeries) -> NDArrayFloat:  # pyright: ignore[reportIncompatibleMethodOverride]
        """
        Transform from Timeddelta to numerical format
        """
        # nanoseconds
        return np.array([_x.value for _x in x])

    @staticmethod
    def inverse(x: FloatArrayLike) -> NDArrayTimedelta:
        """
        Transform to Timedelta from numerical format
        """
        return np.array([pd.Timedelta(int(i)) for i in x])


class reciprocal_trans(trans):
    """
    Reciprocal Transformation
    """

    @staticmethod
    def transform(x: FloatArrayLike) -> NDArrayFloat:
        return 1 / np.asarray(x)

    @staticmethod
    def inverse(x: FloatArrayLike) -> NDArrayFloat:
        return 1 / np.asarray(x)


class pseudo_log_trans(trans):
    """
    Pseudo-log transformation

    A transformation mapping numbers to a signed logarithmic
    scale with a smooth transition to linear scale around 0.

    Parameters
    ----------
    sigma : float
        Scaling factor for the linear part.
    base : int
        Approximate logarithm used. If None, then
        the natural log is used.
    kwargs : dict
        Keyword arguments passed onto
        :func:`trans_new`. Should not include
        the `transform` or `inverse`.
    """

    def __init__(self, sigma=1, base=None, **kwargs):
        if base is None:
            base = np.exp(1)

        self.sigma = sigma
        self.base = base
        super().__init__(**kwargs)

    def transform(self, x: FloatArrayLike) -> NDArrayFloat:  # pyright: ignore[reportIncompatibleMethodOverride]
        x = np.asarray(x)
        return np.arcsinh(x / (2 * self.sigma)) / np.log(self.base)

    def inverse(self, x: FloatArrayLike) -> NDArrayFloat:  # pyright: ignore[reportIncompatibleMethodOverride]
        x = np.asarray(x)
        return 2 * self.sigma * np.sinh(x * np.log(self.base))

    def minor_breaks(
        self,
        major: FloatArrayLike,
        limits: Optional[TupleFloat2] = None,
        n: Optional[int] = None,
    ) -> NDArrayFloat:
        n = int(self.base) - 2 if n is None else n
        return super().minor_breaks(major, limits, n)


class symlog_trans(trans):
    """
    Symmetric Log Transformation

    They symmetric logarithmic transformation is defined as

    ::

        f(x) = log(x+1) for x >= 0
               -log(-x+1) for x < 0

    It can be useful for data that has a wide range of both positive
    and negative values (including zero).
    """

    breaks_: BreaksFunction = breaks_symlog()

    @staticmethod
    def transform(x: FloatArrayLike) -> NDArrayFloat:
        return np.sign(x) * np.log1p(np.abs(x))

    @staticmethod
    def inverse(x: FloatArrayLike) -> NDArrayFloat:
        return np.sign(x) * (np.exp(np.abs(x)) - 1)  # type: ignore


def gettrans(
    t: str | Callable[[], Type[trans]] | Type[trans] | trans | None = None,
):
    """
    Return a trans object

    Parameters
    ----------
    t : str | callable | type | trans
        Name of transformation function. If None, returns an
        identity transform.

    Returns
    -------
    out : trans
    """
    obj = t
    # Make sure trans object is instantiated
    if t is None:
        return identity_trans()

    if isinstance(obj, str):
        name = "{}_trans".format(obj)
        obj = globals()[name]()
    if callable(obj):
        obj = obj()
    if isinstance(obj, type):
        obj = obj()

    if not isinstance(obj, trans):
        raise ValueError("Could not get transform object.")

    return obj



from datetime import datetime

from mizani._core.date_utils import (
    ceil_month,
    expand_datetime_limits,
    shift_limits_down,
)


def test_shift_limits_down():
    lo = (1973, 1998)
    lc = (1973, 2023)
    assert shift_limits_down(lc, lo, 10) == (1970, 2020)

    lo = (1973, 2021)
    lc = (1973, 2023)
    assert shift_limits_down(lc, lo, 10) == lc


def test_expand_datetime_limits():
    limits = (
        datetime(2000, 1, 1, microsecond=100),
        datetime(2000, 1, 1, microsecond=1300),
    )
    assert expand_datetime_limits(limits, 200, "microsecond") == limits


def test_ceil_month():
    d = datetime(2020, 1, 11)
    assert ceil_month(d) == datetime(2020, 2, 1)

    d = datetime(2020, 1, 1)
    assert ceil_month(d) == d



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from ..hsluv import rgb_to_hex
from ._colormap import ColorMap, ColorMapKind

if TYPE_CHECKING:
    from typing import Sequence

    from mizani.typing import (
        FloatArrayLike,
        RGBHexColor,
        SegmentFunctionColorMapData,
    )


@dataclass
class SegmentFunctionMap(ColorMap):
    """
    Gradient colormap by calculating RGB colors independently

    The input data is the same as Matplotlib's LinearSegmentedColormap
    data whose values for each channel are functions.
    """

    data: SegmentFunctionColorMapData
    kind: ColorMapKind = ColorMapKind.miscellaneous

    def _generate_colors(self, x: FloatArrayLike) -> Sequence[RGBHexColor]:
        x = np.asarray(x)
        arr = np.column_stack(
            [
                self.data["red"](x),
                self.data["blue"](x),
                self.data["green"](x),
            ]
        )
        return [rgb_to_hex(c) for c in arr]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ._interpolated import InterpolatedMap

if TYPE_CHECKING:
    from mizani._colors._palettes import palette


class PaletteInterpolatedMap(InterpolatedMap):
    """
    Make a colormap by interpolating the colors of palette
    """

    def __init__(self, palette: palette):
        colors = np.asarray(palette.swatches[-1], dtype=float) / 255
        super().__init__(colors, kind=palette.kind)
        self.palette = palette

    def discrete_palette(self, n):
        """
        Pick exact colors from the swatch if possible
        """
        if n <= self.palette.max_colors:
            return self.palette.get_hex_swatch(n)
        return super().discrete_palette(n)



from __future__ import annotations

import math
from datetime import datetime, timedelta, tzinfo
from typing import TYPE_CHECKING, overload
from zoneinfo import ZoneInfo

import numpy as np
from dateutil.rrule import rrule

from ..utils import get_timezone, isclose_abs
from .date_utils import Interval, align_limits, expand_datetime_limits
from .types import DateFrequency, date_breaks_info

if TYPE_CHECKING:
    from typing import Generator, Optional, Sequence

    from mizani.typing import (
        Datetime,
        DatetimeBreaksUnits,
        FloatArrayLike,
        NDArrayDatetime,
        NDArrayFloat,
        SeqDatetime,
        TzInfo,
    )


EPOCH = datetime(1970, 1, 1, tzinfo=None)
EPOCH64 = np.datetime64("1970", "Y")
SECONDS_PER_DAY = 24 * 60 * 60
MICROSECONDS_PER_DAY = SECONDS_PER_DAY * (10**6)

NaT_int = np.datetime64("NaT").astype(np.int64)
MIN_DATETIME64 = np.datetime64("0001-01-01")
MAX_DATETIME64 = np.datetime64("10000-01-01")
UTC = ZoneInfo("UTC")
DF = DateFrequency


def _from_ordinalf(x: float, tz: tzinfo | None) -> datetime:
    """
    Convert float array to datetime
    """
    dt64 = EPOCH64 + np.timedelta64(
        int(np.round(x * MICROSECONDS_PER_DAY)), "us"
    )
    if not (MIN_DATETIME64 < dt64 <= MAX_DATETIME64):
        raise ValueError(
            f"Date ordinal {x} converts to {dt64} (using "
            f"epoch {EPOCH}). The supported dates must be  "
            "between year 0001 and 9999."
        )

    # convert from datetime64 to datetime:
    dt: datetime = dt64.astype(object)

    # but maybe we are working in a different timezone so move.
    if tz:
        # datetime64 is always UTC:
        dt = dt.replace(tzinfo=UTC)
        dt = dt.astimezone(tz)

    # fix round off errors
    if np.abs(x) > 70 * 365:
        # if x is big, round off to nearest twenty microseconds.
        # This avoids floating point roundoff error
        ms = round(dt.microsecond / 20) * 20
        if ms == 1000000:
            dt = dt.replace(microsecond=0) + timedelta(seconds=1)
        else:
            dt = dt.replace(microsecond=ms)

    return dt


_from_ordinalf_np_vectorized = np.vectorize(_from_ordinalf, otypes="O")


def get_tzinfo(tz: Optional[str | TzInfo] = None) -> TzInfo | None:
    """
    Generate `~datetime.tzinfo` from a string or return `~datetime.tzinfo`.

    If argument is None, return None.
    """
    if tz is None:
        return None

    if isinstance(tz, str):
        return ZoneInfo(tz)

    if isinstance(tz, tzinfo):
        return tz

    raise TypeError("tz must be string or tzinfo subclass.")


@overload
def datetime_to_num(x: SeqDatetime) -> NDArrayFloat: ...


@overload
def datetime_to_num(x: Datetime) -> float: ...


def datetime_to_num(x: SeqDatetime | Datetime) -> NDArrayFloat | float:
    """
    Convery any datetime sequence to float array
    """
    iterable = np.iterable(x)
    _x = x if iterable else [x]
    try:
        x0 = next(iter(_x))
    except StopIteration:
        return np.array([], dtype=float)

    if isinstance(x0, datetime) and x0.tzinfo:
        _x = [
            dt.astimezone(UTC).replace(tzinfo=None)  # type: ignore
            for dt in _x
        ]
    res = datetime64_to_num(np.asarray(_x, dtype="datetime64"))
    return res if iterable else res[0]


def datetime64_to_num(x: NDArrayDatetime) -> NDArrayFloat:
    """
    Convery any numpy datetime64 array to float array
    """
    x_secs = x.astype("datetime64[s]")
    diff_ns = (x - x_secs).astype("timedelta64[ns]")

    # In seconds + nanoseconds, in days
    res = (
        (x_secs - EPOCH64).astype(np.float64)
        + diff_ns.astype(np.float64) / 1.0e9
    ) / SECONDS_PER_DAY

    x_int = x.astype(np.int64)
    res[x_int == NaT_int] = np.nan
    return res


def num_to_datetime(
    x: FloatArrayLike, tz: Optional[str | TzInfo] = None
) -> NDArrayDatetime:
    """
    Convert any float array to numpy datetime64 array
    """
    tz = get_tzinfo(tz) or UTC
    return _from_ordinalf_np_vectorized(x, tz)


WIDTHS: dict[DateFrequency, Sequence[int]] = {
    DF.YEARLY: (1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000),
    DF.MONTHLY: (1, 2, 3, 4, 6),
    DF.DAILY: (1, 2, 4, 7, 14),
    DF.HOURLY: (1, 2, 3, 4, 6, 12),
    DF.MINUTELY: (1, 5, 10, 15, 30),
    DF.SECONDLY: (1, 5, 10, 15, 30),
    DF.MICROSECONDLY: (
        1,
        2,
        5,
        10,
        20,
        50,
        100,
        200,
        500,
        1000,
        2000,
        5000,
        10000,
        20000,
        50000,
        100000,
        200000,
        500000,
        1000000,
    ),
}

MAX_BREAKS = dict(zip(DateFrequency, (11, 12, 11, 12, 11, 11, 8)))


def _viable_freqs(
    min_breaks: int,
    unit_durations: tuple[int, int, int, int, int, int, int],
) -> Generator[tuple[DateFrequency, int], None, None]:
    """
    Find viable frequency, duration pairs

    A pair is viable if it can yeild a suitable number of breaks
    For example:
        - YEARLY frequency, 3 year unit_duration and
          8 min_breaks is not viable
        - MONTHLY frequency, 36 month unit_duration and
          8 min_breaks is viable
    """
    for freq, duration in zip(DateFrequency, unit_durations):
        max_width = WIDTHS[freq][-1]
        max_breaks = max(min_breaks, MAX_BREAKS.get(freq, 11))
        if duration <= max_width * max_breaks - 1:
            yield (freq, duration)


def calculate_date_breaks_info(
    limits: tuple[datetime, datetime], n: int = 5
) -> date_breaks_info:
    """
    Calculate information required to generate breaks

    Parameters
    ----------
    limits:
        Datetime limits for the breaks
    n:
        Desired number of breaks.
    """
    tz = get_timezone(limits)
    _max_breaks_lookup = {f: max(b, n) for f, b in MAX_BREAKS.items()}

    # Widen the duration at each granularity
    itv = Interval(*limits)
    unit_durations = (
        itv.y_wide,
        itv.M_wide,
        itv.d_wide,
        itv.h_wide,
        itv.m_wide,
        itv.s,
        itv.u,
    )
    # Search frequencies from longest (yearly) to the smallest
    # for one that would a good width between the breaks

    # Defaults
    freq = DF.YEARLY  # makes pyright happy
    break_width = 1
    duration = n

    for freq, duration in _viable_freqs(n, unit_durations):
        # Search for breaks in the range
        _max_breaks = _max_breaks_lookup[freq]
        for mb in range(n, _max_breaks + 1):
            # There are few breaks at this frequency
            # e.g. (freq=YEARLY, duration=2) but mb = 5
            if duration < mb:
                continue

            for break_width in WIDTHS[freq]:
                if duration <= break_width * mb - 1:
                    break
            else:  # pragma: no cover
                continue
            break
        else:
            continue
        break

    num_breaks = duration // break_width
    limits = itv.limits_for_frequency(freq)
    res = date_breaks_info(
        freq,
        num_breaks,
        break_width,
        start=limits[0].replace(tzinfo=tz),
        until=limits[1].replace(tzinfo=tz),
        tz=tz,
    )
    return res


def calculate_date_breaks_auto(limits, n: int = 5) -> Sequence[datetime]:
    """
    Calcuate date breaks using appropriate units
    """
    info = calculate_date_breaks_info(limits, n=n)
    lookup = {
        DF.YEARLY: yearly_breaks,
        DF.MONTHLY: monthly_breaks,
        DF.DAILY: daily_breaks,
        DF.HOURLY: hourly_breaks,
        DF.MINUTELY: minutely_breaks,
        DF.SECONDLY: secondly_breaks,
        DF.MICROSECONDLY: microsecondly_breaks,
    }
    return lookup[info.frequency](info)


def calculate_date_breaks_byunits(
    limits,
    units: DatetimeBreaksUnits,
    width: int,
    max_breaks: Optional[int] = None,
    tz: Optional[TzInfo] = None,
) -> Sequence[datetime]:
    """
    Calcuate date breaks using appropriate units
    """
    timely_name = f"{units.upper()}LY"
    if timely_name in ("DAYLY", "WEEKLY"):
        timely_name = "DAILY"

    freq = getattr(DF, timely_name)

    # Appropriate start and end dates
    start, until = expand_datetime_limits(limits, width, units)

    if units == "week":
        width *= 7

    info = date_breaks_info(
        freq,
        n=-1,
        width=width,
        start=start,
        until=until,
        tz=tz,
    )

    lookup = {
        "year": rrulely_breaks,
        "month": rrulely_breaks,
        "week": rrulely_breaks,
        "day": rrulely_breaks,
        "hour": rrulely_breaks,
        "minute": rrulely_breaks,
        "second": rrulely_breaks,
        "microsecond": microsecondly_breaks,
    }
    return lookup[units](info)


def rrulely_breaks(info: date_breaks_info) -> Sequence[datetime]:
    r = rrule(
        info.frequency,
        interval=info.width,
        dtstart=info.start,
        until=info.until,
    )
    return list(r)


def yearly_breaks(info: date_breaks_info) -> Sequence[datetime]:
    """
    Calculate yearly breaks
    """
    # New limits so that breaks fall on multiples of
    # the width
    limits = info.start.year, info.until.year
    l, h = align_limits(limits, info.width)
    l, h = math.floor(l), math.ceil(h)

    _replace_d = {
        "month": 1,
        "day": 1,
        "hour": 0,
        "minute": 0,
        "second": 0,
        "tzinfo": info.tz,
    }

    start = info.start.replace(year=l, **_replace_d)
    until = info.until.replace(year=h, **_replace_d)
    r = rrule(
        info.frequency,
        interval=info.width,
        dtstart=start,
        until=until,
    )
    return list(r)


def monthly_breaks(info: date_breaks_info) -> Sequence[datetime]:
    """
    Calculate monthly breaks
    """
    r = rrule(
        info.frequency,
        interval=info.width,
        dtstart=info.start,
        until=info.until,
    )
    return list(r)


def daily_breaks(info: date_breaks_info) -> Sequence[datetime]:
    """
    Calculate daily breaks
    """
    if info.width == 7:
        interval = 1
        bymonthday = (1, 8, 15, 22)
    elif info.width == 14:
        interval = 1
        bymonthday = (1, 15)
    else:
        interval = info.width
        bymonthday = None

    r = rrule(
        info.frequency,
        interval=interval,
        dtstart=info.start,
        until=info.until,
        bymonthday=bymonthday,
    )
    return r.between(info.start, info.until, True)


def hourly_breaks(info: date_breaks_info) -> Sequence[datetime]:
    """
    Calculate hourly breaks
    """
    r = rrule(
        info.frequency,
        interval=1,
        dtstart=info.start,
        until=info.until,
        byhour=range(0, 24, info.width),
    )
    return r.between(info.start, info.until, True)


def minutely_breaks(info: date_breaks_info) -> Sequence[datetime]:
    """
    Calculate minutely breaks
    """
    r = rrule(
        info.frequency,
        interval=1,
        dtstart=info.start,
        until=info.until,
        byminute=range(0, 60, info.width),
    )
    return r.between(info.start, info.until, True)


def secondly_breaks(info: date_breaks_info) -> Sequence[datetime]:
    """
    Calculate minutely breaks
    """
    r = rrule(
        info.frequency,
        interval=1,
        dtstart=info.start,
        until=info.until,
        bysecond=range(0, 60, info.width),
    )
    return r.between(info.start, info.until, True)


def microsecondly_breaks(info: date_breaks_info) -> NDArrayDatetime:
    """
    Calculate breaks at microsecond intervals
    """
    # TODO: A little too complicated, could use some refactoring
    nmin: float
    nmax: float
    width = info.width

    nmin, nmax = datetime_to_num((info.start, info.until))
    day0: float = np.floor(nmin)

    # difference in microseconds
    umax = (nmax - day0) * MICROSECONDS_PER_DAY
    umin = (nmin - day0) * MICROSECONDS_PER_DAY

    # Ensure max is a multiple of the width
    width = info.width
    h, m = divmod(umax, width)
    if not isclose_abs(m / width, 0):
        h += 1
    umax = h * width

    # Generate breaks at the multiples of the width
    n = (umax - umin + 0.001 * width) // width
    ubreaks = umin - width + np.arange(n + 3) * width
    breaks = day0 + ubreaks / MICROSECONDS_PER_DAY
    return num_to_datetime(breaks, info.tz)



import pytest

from mizani._colors import InterpolatedMap


def test_number_of_values():
    with pytest.raises(ValueError):
        InterpolatedMap(["blue", "red", "green"], [0.0, 0.2, 0.9, 1.0])

    with pytest.raises(ValueError):
        InterpolatedMap(["blue", "red", "green"], [0, 0.2, 0.9])

    with pytest.raises(ValueError):
        InterpolatedMap(["blue"], [0])


def test_discrete():
    gmap = InterpolatedMap(["blue", "red", "green"], [0, 0.5, 1])
    colors = gmap.discrete_palette(5)
    assert len(colors) == 5
    assert [c is not None and c.startswith("#") for c in colors]



from ._colormaps._colormap import ColorMap, ColorMapKind
from ._colormaps._cubehelix import CubeHelixMap
from ._colormaps._interpolated import InterpolatedMap
from ._colormaps._listed import ListedMap
from ._colormaps._segment_function import SegmentFunctionMap
from ._colormaps._segment_interpolated import SegmentInterpolatedMap
from .hsluv import hex_to_rgb, rgb_to_hex
from .named_colors import get_colormap, get_named_color

__all__ = (
    "ColorMap",
    "ColorMapKind",
    "CubeHelixMap",
    "InterpolatedMap",
    "ListedMap",
    "SegmentFunctionMap",
    "SegmentInterpolatedMap",
    "get_colormap",
    "get_named_color",
    "hex_to_rgb",
    "rgb_to_hex",
)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from mizani._colors import SegmentFunctionMap

if TYPE_CHECKING:
    from typing import Callable

    from mizani.typing import (
        NDArrayFloat,
        SegmentFunctionColorMapData,
    )

__all__ = (
    # miscellaneous
    "afmhot",
    "flag",
    "gist_heat",
    "gnuplot",
    "gnuplot2",
    "ocean",
    "prism",
    "rainbow",
)


# Gnuplot palette functions
def _gpf_32(x: NDArrayFloat) -> NDArrayFloat:
    ret = np.zeros(len(x))
    m = x < 0.25
    ret[m] = 4 * x[m]
    m = (x >= 0.25) & (x < 0.92)
    ret[m] = -2 * x[m] + 1.84
    m = x >= 0.92
    ret[m] = x[m] / 0.08 - 11.5
    return ret


GPF: dict[int, Callable[[NDArrayFloat], NDArrayFloat]] = {
    0: lambda x: np.zeros(len(x)),
    1: lambda x: np.full(len(x), 0.5),
    2: lambda x: np.ones(len(x)),
    3: lambda x: x,
    4: lambda x: x**2,
    5: lambda x: x**3,
    6: lambda x: x**4,
    7: lambda x: np.sqrt(x),
    8: lambda x: np.sqrt(np.sqrt(x)),
    9: lambda x: np.sin(x * np.pi / 2),
    10: lambda x: np.cos(x * np.pi / 2),
    11: lambda x: np.abs(x - 0.5),
    12: lambda x: (2 * x - 1) ** 2,
    13: lambda x: np.sin(x * np.pi),
    14: lambda x: np.abs(np.cos(x * np.pi)),
    15: lambda x: np.sin(x * 2 * np.pi),
    16: lambda x: np.cos(x * 2 * np.pi),
    17: lambda x: np.abs(np.sin(x * 2 * np.pi)),
    18: lambda x: np.abs(np.cos(x * 2 * np.pi)),
    19: lambda x: np.abs(np.sin(x * 4 * np.pi)),
    20: lambda x: np.abs(np.cos(x * 4 * np.pi)),
    21: lambda x: 3 * x,
    22: lambda x: 3 * x - 1,
    23: lambda x: 3 * x - 2,
    24: lambda x: np.abs(3 * x - 1),
    25: lambda x: np.abs(3 * x - 2),
    26: lambda x: (3 * x - 1) / 2,
    27: lambda x: (3 * x - 2) / 2,
    28: lambda x: np.abs((3 * x - 1) / 2),
    29: lambda x: np.abs((3 * x - 2) / 2),
    30: lambda x: x / 0.32 - 0.78125,
    31: lambda x: 2 * x - 0.84,
    32: _gpf_32,
    33: lambda x: np.abs(2 * x - 0.5),
    34: lambda x: 2 * x,
    35: lambda x: 2 * x - 0.5,
    36: lambda x: 2 * x - 1,
}

_afmhot: SegmentFunctionColorMapData = {
    "red": GPF[34],
    "green": GPF[35],
    "blue": GPF[36],
}

_flag: SegmentFunctionColorMapData = {
    "red": lambda x: 0.75 * np.sin((x * 31.5 + 0.25) * np.pi) + 0.5,
    "blue": lambda x: np.sin(x * 31.5 * np.pi),
    "green": lambda x: 0.75 * np.sin((x * 31.5 - 0.25) * np.pi) + 0.5,
}

_gist_heat: SegmentFunctionColorMapData = {
    "red": lambda x: 1.5 * x,
    "green": lambda x: 2 * x - 1,
    "blue": lambda x: 4 * x - 3,
}

_gnuplot: SegmentFunctionColorMapData = {
    "red": GPF[7],
    "green": GPF[5],
    "blue": GPF[15],
}

_gnuplot2: SegmentFunctionColorMapData = {
    "red": GPF[30],
    "green": GPF[31],
    "blue": GPF[32],
}

_ocean: SegmentFunctionColorMapData = {
    "red": GPF[23],
    "green": GPF[28],
    "blue": GPF[3],
}

_prism: SegmentFunctionColorMapData = {
    "red": lambda x: 0.75 * np.sin((x * 20.9 + 0.25) * np.pi) + 0.67,
    "blue": lambda x: 0.75 * np.sin((x * 20.9 - 0.25) * np.pi) + 0.33,
    "green": lambda x: -1.1 * np.sin((x * 20.9) * np.pi),
}

_rainbow: SegmentFunctionColorMapData = {
    "red": GPF[33],
    "green": GPF[13],
    "blue": GPF[10],
}

afmhot = SegmentFunctionMap(_afmhot)
flag = SegmentFunctionMap(_flag)
gist_heat = SegmentFunctionMap(_gist_heat)
gnuplot = SegmentFunctionMap(_gnuplot)
gnuplot2 = SegmentFunctionMap(_gnuplot2)
ocean = SegmentFunctionMap(_ocean)
prism = SegmentFunctionMap(_prism)
rainbow = SegmentFunctionMap(_rainbow)



import numpy as np

from mizani._colors import ListedMap


def test_continuous():
    gmap = ListedMap(["blue", "red", "green"])
    x = np.linspace(0, 1, 10)
    colors = gmap.continuous_palette(x)
    assert len(colors) == 10
    assert [c is not None and c.startswith("#") for c in colors]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from ..hsluv import hex_to_rgb, rgb_to_hex
from ._colormap import ColorMap, ColorMapKind

if TYPE_CHECKING:
    from typing import Sequence

    from mizani.typing import (
        FloatArrayLike,
        RGBColor,
        RGBColorArray,
        RGBHexColor,
    )


SPACE256 = np.arange(256)
INNER_SPACE256 = SPACE256[1:-1]
ROUNDING_JITTER = 1e-12

__all__ = ("ListedMap",)


@dataclass
class ListedMap(ColorMap):
    colors: Sequence[RGBHexColor] | Sequence[RGBColor] | RGBColorArray
    kind: ColorMapKind = ColorMapKind.miscellaneous

    def __post_init__(self):
        from ..named_colors import get_named_color

        self.values = np.linspace(0, 1, len(self.colors))
        if isinstance(self.colors[0], str):
            colors = [
                hex_to_rgb(get_named_color(c))  # type: ignore
                for c in self.colors
            ]
        else:
            colors = self.colors

        self.n = len(colors)
        self._data = np.asarray(colors)

    def _generate_colors(self, x: FloatArrayLike) -> Sequence[RGBHexColor]:
        """
        Lookup colors in the interpolated ranges

        Parameters
        ----------
        x :
            Values in the range [0, 1]. O maps to the start of the
            gradient, and 1 to the end of the gradient.
        """
        x = np.asarray(x)
        idx = (x * self.n).astype(int)
        idx[idx < 0] = 0
        idx[idx >= self.n] = self.n - 1
        arr = self._data.take(idx, axis=0, mode="clip")
        return [rgb_to_hex(c) for c in arr]



from ._colormap import ColorMap
from ._cubehelix import CubeHelixMap
from ._interpolated import InterpolatedMap
from ._listed import ListedMap
from ._palette_interpolated import PaletteInterpolatedMap
from ._segment_function import SegmentFunctionMap
from ._segment_interpolated import SegmentInterpolatedMap

__all__ = (
    "ColorMap",
    "CubeHelixMap",
    "InterpolatedMap",
    "ListedMap",
    "PaletteInterpolatedMap",
    "SegmentFunctionMap",
    "SegmentInterpolatedMap",
)



"""
Palettes are the link between data values and the values along
the dimension of a scale. Before a collection of values can be
represented on a scale, they are transformed by a palette. This
transformation is knowing as mapping. Values are mapped onto a
scale by a palette.

Scales tend to have restrictions on the magnitude of quantities
that they can intelligibly represent. For example, the size of
a point should be significantly smaller than the plot panel
onto which it is plotted or else it would be hard to compare
two or more points. Therefore palettes must be created that
enforce such restrictions. This is the reason for the ``*_pal``
functions that create and return the actual palette functions.
"""

from __future__ import annotations

import colorsys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol
from warnings import warn

import numpy as np

from ._colors import (
    CubeHelixMap,
    InterpolatedMap,
    get_colormap,
    get_named_color,
    hex_to_rgb,
    hsluv,
    rgb_to_hex,
)
from .bounds import rescale
from .utils import identity

if TYPE_CHECKING:
    from typing import Any, Literal, Optional, Sequence, TypeVar

    from mizani.typing import (
        Callable,
        ColorScheme,
        ColorSchemeShort,
        FloatArrayLike,
        NDArrayFloat,
        RGBHexColor,
        TupleFloat2,
        TupleFloat3,
    )

    T = TypeVar("T")


__all__ = [
    "hls_palette",
    "hsluv_palette",
    "rescale_pal",
    "area_pal",
    "abs_area",
    "grey_pal",
    "hue_pal",
    "brewer_pal",
    "gradient_n_pal",
    "cmap_pal",
    "cmap_d_pal",
    "desaturate_pal",
    "manual_pal",
    "xkcd_palette",
    "crayon_palette",
    "cubehelix_pal",
    "identity_pal",
    "none_pal",
]


class _discrete_pal(Protocol):
    """
    Discrete palette maker
    """

    def __call__(self, n: int) -> Sequence[Any]:
        """
        Palette method
        """
        ...


class _continuous_pal(Protocol):
    """
    Continuous palette maker
    """

    def __call__(self, x: Sequence[Any]) -> NDArrayFloat:
        """
        Palette method
        """
        ...


class _continuous_color_pal(Protocol):
    """
    Continuous color palette maker
    """

    def __call__(self, x: FloatArrayLike) -> Sequence[RGBHexColor | None]:
        """
        Palette method
        """
        ...


def hls_palette(
    n_colors: int = 6, h: float = 0.01, l: float = 0.6, s: float = 0.65
) -> Sequence[TupleFloat3]:
    """
    Get a set of evenly spaced colors in HLS hue space.

    h, l, and s should be between 0 and 1

    Parameters
    ----------

    n_colors : int
        number of colors in the palette
    h : float
        first hue
    l : float
        lightness
    s : float
        saturation

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    See Also
    --------
    hsluv_palette : Make a palette using evenly spaced circular
        hues in the HSLuv system.

    Examples
    --------
    >>> len(hls_palette(2))
    2
    >>> len(hls_palette(9))
    9
    """
    hues = np.linspace(0, 1, n_colors + 1)[:-1]
    hues += h
    hues %= 1
    hues -= hues.astype(int)
    palette = [colorsys.hls_to_rgb(h_i, l, s) for h_i in hues]
    return palette


def hsluv_palette(
    n_colors: int = 6, h: float = 0.01, s: float = 0.9, l: float = 0.65
) -> Sequence[TupleFloat3]:
    """
    Get a set of evenly spaced colors in HSLuv hue space.

    h, s, and l should be between 0 and 1

    Parameters
    ----------

    n_colors : int
        number of colors in the palette
    h : float
        first hue
    s : float
        saturation
    l : float
        lightness

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    See Also
    --------
    hls_palette : Make a palette using evenly spaced circular
        hues in the HSL system.

    Examples
    --------
    >>> len(hsluv_palette(3))
    3
    >>> len(hsluv_palette(11))
    11
    """
    hues = np.linspace(0, 1, n_colors + 1)[:-1]
    hues += h
    hues %= 1
    hues *= 359
    s *= 99
    l *= 99
    palette = [hsluv.hsluv_to_rgb((h_i, s, l)) for h_i in hues]
    return palette


@dataclass
class rescale_pal(_continuous_pal):
    """
    Rescale the input to the specific output range.

    Useful for alpha, size, and continuous position.

    Parameters
    ----------
    range : tuple
        Range of the scale

    Returns
    -------
    out : function
        Palette function that takes a sequence of values
        in the range ``[0, 1]`` and returns values in
        the specified range.

    Examples
    --------
    >>> palette = rescale_pal()
    >>> palette([0, .2, .4, .6, .8, 1])
    array([0.1 , 0.28, 0.46, 0.64, 0.82, 1.  ])

    The returned palette expects inputs in the ``[0, 1]``
    range. Any value outside those limits is clipped to
    ``range[0]`` or ``range[1]``.

    >>> palette([-2, -1, 0.2, .4, .8, 2, 3])
    array([0.1 , 0.1 , 0.28, 0.46, 0.82, 1.  , 1.  ])
    """

    range: TupleFloat2 = (0.1, 1)

    def __call__(self, x: FloatArrayLike) -> NDArrayFloat:
        return rescale(x, self.range, _from=(0, 1))


@dataclass
class area_pal(_continuous_pal):
    """
    Point area palette (continuous).

    Parameters
    ----------
    range : tuple
        Numeric vector of length two, giving range of possible sizes.
        Should be greater than 0.

    Returns
    -------
    out : function
        Palette function that takes a sequence of values
        in the range ``[0, 1]`` and returns values in
        the specified range.

    Examples
    --------
    >>> x = np.arange(0, .6, .1)**2
    >>> palette = area_pal()
    >>> palette(x)
    array([1. , 1.5, 2. , 2.5, 3. , 3.5])

    The results are equidistant because the input ``x`` is in
    area space, i.e it is squared.
    """

    range: TupleFloat2 = (1, 6)

    def __call__(self, x: FloatArrayLike) -> NDArrayFloat:
        return rescale(np.sqrt(x), to=self.range, _from=(0, 1))


@dataclass
class abs_area(_continuous_pal):
    """
    Point area palette (continuous), with area proportional to value.

    Parameters
    ----------
    max : float
        A number representing the maximum size

    Returns
    -------
    out : function
        Palette function that takes a sequence of values
        in the range ``[0, 1]`` and returns values in the range
        ``[0, max]``.

    Examples
    --------
    >>> x = np.arange(0, .8, .1)**2
    >>> palette = abs_area(5)
    >>> palette(x)
    array([0. , 0.5, 1. , 1.5, 2. , 2.5, 3. , 3.5])

    Compared to :func:`area_pal`, :func:`abs_area` will handle values
    in the range ``[-1, 0]`` without returning ``np.nan``. And values
    whose absolute value is greater than 1 will be clipped to the
    maximum.
    """

    max: float

    def __call__(self, x: FloatArrayLike) -> NDArrayFloat:
        return rescale(np.sqrt(np.abs(x)), to=(0, self.max), _from=(0, 1))


@dataclass
class grey_pal(_discrete_pal):
    """
    Utility for creating continuous grey scale palette

    Parameters
    ----------
    start : float
        grey value at low end of palette
    end : float
        grey value at high end of palette

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        equally spaced colors.

    Examples
    --------
    >>> palette = grey_pal()
    >>> palette(5)
    ['#333333', '#737373', '#989898', '#b4b4b4', '#cccccc']
    """

    start: float = 0.2
    end: float = 0.8

    def __post_init__(self):
        start, end = self.start, self.end
        colors = (start, start, start), (end, end, end)
        self._cmap = InterpolatedMap(colors)

    def __call__(self, n: int) -> Sequence[RGBHexColor | None]:
        gamma = 2.2
        # The grey scale points are linearly separated in
        # gamma encoded space
        space = np.linspace(self.start**gamma, self.end**gamma, n)
        # Map points onto the [0, 1] palette domain
        x = (space ** (1.0 / gamma) - self.start) / (self.end - self.start)
        return self._cmap.continuous_palette(x)


@dataclass
class hue_pal(_discrete_pal):
    """
    Utility for making hue palettes for color schemes.

    Parameters
    ----------
    h : float
        first hue. In the [0, 1] range
    l : float
        lightness. In the [0, 1] range
    s : float
        saturation. In the [0, 1] range
    color_space : 'hls' | 'hsluv'
        Color space to use for the palette.
        `hls` for https://en.wikipedia.org/wiki/HSL_and_HSV
        or `hsluv` for https://www.hsluv.org/.

    Returns
    -------
    out : function
        A discrete color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        equally spaced colors. Though the palette
        is continuous, since it is varies the hue it
        is good for categorical data. However if ``n``
        is large enough the colors show continuity.

    Examples
    --------
    >>> hue_pal()(5)
    ['#db5f57', '#b9db57', '#57db94', '#5784db', '#c957db']
    >>> hue_pal(color_space='hsluv')(5)
    ['#e0697e', '#9b9054', '#569d79', '#5b98ab', '#b675d7']
    """

    h: float = 0.01
    l: float = 0.6
    s: float = 0.65
    color_space: Literal["hls", "hsluv"] = "hls"

    def __post_init__(self):
        h, l, s = self.h, self.l, self.s
        if not all(0 <= val <= 1 for val in (h, l, s)):
            msg = (
                "hue_pal expects values to be between 0 and 1. "
                f"I got {h=}, {l=}, {s=}"
            )
            raise ValueError(msg)

        # For backward compatibility
        if self.color_space == "husl":
            self.color_space = "hsluv"

        if self.color_space not in ("hls", "hsluv"):
            msg = "color_space should be one of ['hls', 'hsluv']"
            raise ValueError(msg)

    def __call__(self, n: int) -> Sequence[RGBHexColor]:
        lookup = {"hls": hls_palette, "hsluv": hsluv_palette}
        palette = lookup[self.color_space]
        colors = palette(n, h=self.h, l=self.l, s=self.s)
        return [hsluv.rgb_to_hex(c) for c in colors]


@dataclass
class brewer_pal(_discrete_pal):
    """
    Utility for making a brewer palette

    Parameters
    ----------
    type : 'sequential' | 'qualitative' | 'diverging'
        Type of palette. Sequential, Qualitative or
        Diverging. The following abbreviations may
        be used, ``seq``, ``qual`` or ``div``.
    palette : int | str
        Which palette to choose from. If is an integer,
        it must be in the range ``[0, m]``, where ``m``
        depends on the number sequential, qualitative or
        diverging palettes. If it is a string, then it
        is the name of the palette.
    direction : int
        The order of colours in the scale. If -1 the order
        of colors is reversed. The default is 1.

    Returns
    -------
    out : function
        A color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        colors. The maximum value of ``n`` varies
        depending on the parameters.

    Examples
    --------
    >>> brewer_pal()(5)
    ['#EFF3FF', '#BDD7E7', '#6BAED6', '#3182BD', '#08519C']
    >>> brewer_pal('qual')(5)
    ['#7FC97F', '#BEAED4', '#FDC086', '#FFFF99', '#386CB0']
    >>> brewer_pal('qual', 2)(5)
    ['#1B9E77', '#D95F02', '#7570B3', '#E7298A', '#66A61E']
    >>> brewer_pal('seq', 'PuBuGn')(5)
    ['#F6EFF7', '#BDC9E1', '#67A9CF', '#1C9099', '#016C59']

    The available color names for each palette type can be
    obtained using the following code::

        from mizani._colors.brewer import get_palette_names

        print(get_palette_names("sequential"))
        print(get_palette_names("qualitative"))
        print(get_palette_names("diverging"))
    """

    type: ColorScheme | ColorSchemeShort = "seq"
    palette: int | str = 1
    direction: Literal[1, -1] = 1

    def __post_init__(self):
        from mizani._colors._palettes.brewer import get_brewer_palette

        if self.direction not in (1, -1):
            raise ValueError("direction should be 1 or -1.")

        self.bpal = get_brewer_palette(self.type, self.palette)

    def __call__(self, n: int) -> Sequence[RGBHexColor | None]:
        # Only draw the maximum allowable colors from the palette
        # and fill any remaining spots with None
        _n = min(max(n, self.bpal.min_colors), self.bpal.max_colors)
        color_map = self.bpal.get_hex_swatch(_n)
        colors = color_map[:n]
        if n > self.bpal.max_colors:
            msg = (
                "Warning message:"
                f"Brewer palette {self.bpal.name} has a maximum "
                f"of {self.bpal.max_colors} colors Returning the "
                "palette you asked for with that many colors"
            )
            warn(msg)
            colors = list(colors) + [None] * (n - self.bpal.max_colors)
        return colors[:: self.direction]


@dataclass
class gradient_n_pal(_continuous_color_pal):
    """
    Create a n color gradient palette

    Parameters
    ----------
    colors : list
        list of colors
    values : list, optional
        list of points in the range [0, 1] at which to
        place each color. Must be the same size as
        `colors`. Default to evenly space the colors

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        parameter either a :class:`float` or a sequence
        of floats maps those value(s) onto the palette
        and returns color(s). The float(s) must be
        in the range [0, 1].

    Examples
    --------
    >>> palette = gradient_n_pal(['red', 'blue'])
    >>> palette([0, .25, .5, .75, 1])
    ['#ff0000', '#bf0040', '#7f0080', '#4000bf', '#0000ff']
    >>> palette([-np.inf, 0, np.nan, 1, np.inf])
    [None, '#ff0000', None, '#0000ff', None]
    """

    colors: Sequence[str]
    values: Optional[Sequence[float]] = None

    def __post_init__(self):
        self._cmap = InterpolatedMap(self.colors, self.values)

    def __call__(self, x: FloatArrayLike) -> Sequence[RGBHexColor | None]:
        return self._cmap.continuous_palette(x)


@dataclass
class cmap_pal(_continuous_color_pal):
    """
    Create a continuous palette using a colormap

    Parameters
    ----------
    name : str
        Name of colormap

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        parameter either a :class:`float` or a sequence
        of floats maps those value(s) onto the palette
        and returns color(s). The float(s) must be
        in the range [0, 1].

    Examples
    --------
    >>> palette = cmap_pal('viridis')
    >>> palette([.1, .2, .3, .4, .5])
    ['#482475', '#414487', '#355f8d', '#2a788e', '#21918c']
    """

    name: str

    def __post_init__(self):
        self._direction: Literal[1, -1] = 1
        # Accomodate matplotlib naming convention and allow
        # names that end with _r to return reversed colormaps
        if self.name.endswith("_r"):
            self._direction = -1
            self.name = self.name[:-2]
        self.cm = get_colormap(self.name)

    def __call__(self, x: FloatArrayLike) -> Sequence[RGBHexColor | None]:
        if self._direction == -1:
            x = 1.0 - np.asarray(x)
        return self.cm.continuous_palette(x)


@dataclass
class cmap_d_pal(_discrete_pal):
    """
    Create a discrete palette from a colormap

    Parameters
    ----------
    name : str
        Name of colormap

    Returns
    -------
    out : function
        A discrete color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        colors. The maximum value of ``n`` varies
        depending on the parameters.

    Examples
    --------
    >>> palette = cmap_d_pal('viridis')
    >>> palette(5)
    ['#440154', '#3b528b', '#21918c', '#5ec962', '#fde725']
    """

    name: str

    def __post_init__(self):
        self._direction: Literal[1, -1] = 1
        # Accomodate matplotlib naming convention and allow
        # names that end with _r to return reversed colormaps
        if self.name.endswith("_r"):
            self._direction = -1
            self.name = self.name[:-2]
        self.cm = get_colormap(self.name)

    def __call__(self, n: int) -> Sequence[RGBHexColor]:
        return self.cm.discrete_palette(n)[:: self._direction]


class desaturate_pal(gradient_n_pal):
    """
    Create a palette that desaturate a color by some proportion

    Parameters
    ----------
    color : color
        html color name, hex, rgb-tuple
    prop : float
        saturation channel of color will be multiplied by
        this value
    reverse : bool
        Whether to reverse the palette.

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        parameter either a :class:`float` or a sequence
        of floats maps those value(s) onto the palette
        and returns color(s). The float(s) must be
        in the range [0, 1].

    Examples
    --------
    >>> palette = desaturate_pal('red', .1)
    >>> palette([0, .25, .5, .75, 1])
    ['#ff0000', '#e21d1d', '#c53a3a', '#a95656', '#8c7373']
    """

    def __init__(self, color: str, prop: float, reverse: bool = False):
        if not 0 <= prop <= 1:
            raise ValueError("prop must be between 0 and 1")

        if isinstance(color, str):
            color = get_named_color(color)

        rgb = hex_to_rgb(color)
        h, l, s = colorsys.rgb_to_hls(*rgb)
        s *= prop
        desaturated_color = rgb_to_hex(colorsys.hls_to_rgb(h, l, s))
        colors = [color, desaturated_color]
        if reverse:
            colors = colors[::-1]

        super().__init__(colors)


@dataclass
class manual_pal(_discrete_pal):
    """
    Create a palette from a list of values

    Parameters
    ----------
    values : sequence
        Values that will be returned by the palette function.

    Returns
    -------
    out : function
        A function palette that takes a single
        :class:`int` parameter ``n`` and returns ``n`` values.

    Examples
    --------
    >>> palette = manual_pal(['a', 'b', 'c', 'd', 'e'])
    >>> palette(3)
    ['a', 'b', 'c']
    """

    values: Sequence[Any]

    def __post_init__(self):
        self.size = len(self.values)

    def __call__(self, n: int) -> Sequence[Any]:
        if n > self.size:
            warn(
                f"Palette can return a maximum of {self.size} values. "
                f"{n} values requested."
            )
        return self.values[:n]


def xkcd_palette(colors: Sequence[str]) -> Sequence[RGBHexColor]:
    """
    Make a palette with color names from the xkcd color survey.

    See xkcd for the full list of colors: http://xkcd.com/color/rgb/

    Parameters
    ----------
    colors : list of strings
        List of keys in the ``mizani.colors.xkcd_rgb`` dictionary.

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    Examples
    --------
    >>> palette = xkcd_palette(['red', 'green', 'blue'])
    >>> palette
    ['#E50000', '#15B01A', '#0343DF']

    >>> from mizani._colors.named_colors import XKCD
    >>> list(sorted(XKCD.keys()))[:4]
    ['xkcd:acid green', 'xkcd:adobe', 'xkcd:algae', 'xkcd:algae green']
    """
    return [get_named_color(f"xkcd:{name}") for name in colors]


def crayon_palette(colors: Sequence[str]) -> Sequence[RGBHexColor]:
    """
    Make a palette with color names from Crayola crayons.

    The colors come from
    http://en.wikipedia.org/wiki/List_of_Crayola_crayon_colors

    Parameters
    ----------
    colors : list of strings
        List of keys in the ``mizani.colors.crayloax_rgb`` dictionary.

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    Examples
    --------
    >>> palette = crayon_palette(['almond', 'silver', 'yellow'])
    >>> palette
    ['#EED9C4', '#C9C0BB', '#FBE870']

    >>> from mizani._colors.named_colors import CRAYON
    >>> list(sorted(CRAYON.keys()))[:3]
    ['crayon:almond', 'crayon:antique brass', 'crayon:apricot']
    """
    return [get_named_color(f"crayon:{name}") for name in colors]


@dataclass
class cubehelix_pal(_discrete_pal):
    """
    Utility for creating discrete palette from the cubehelix system.

    This produces a colormap with linearly-decreasing (or increasing)
    brightness. That means that information will be preserved if printed to
    black and white or viewed by someone who is colorblind.

    Parameters
    ----------
    start : float (0 <= start <= 3)
        The hue at the start of the helix.
    rot : float
        Rotations around the hue wheel over the range of the palette.
    gamma : float (0 <= gamma)
        Gamma factor to emphasize darker (gamma < 1) or lighter (gamma > 1)
        colors.
    hue : float (0 <= hue <= 1)
        Saturation of the colors.
    dark : float (0 <= dark <= 1)
        Intensity of the darkest color in the palette.
    light : float (0 <= light <= 1)
        Intensity of the lightest color in the palette.
    reverse : bool
        If True, the palette will go from dark to light.

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        equally spaced colors.


    References
    ----------
    Green, D. A. (2011). "A colour scheme for the display of astronomical
    intensity images". Bulletin of the Astromical Society of India, Vol. 39,
    p. 289-295.

    Examples
    --------
    >>> palette = cubehelix_pal()
    >>> palette(5)
    ['#edd1cb', '#d499a7', '#aa678f', '#6e4071', '#2d1e3e']
    """

    start: int = 0
    rotation: float = 0.4
    gamma: float = 1.0
    hue: float = 0.8
    light: float = 0.85
    dark: float = 0.15
    reverse: bool = False

    def __post_init__(self):
        self._chmap = CubeHelixMap(
            self.start,
            self.rotation,
            self.gamma,
            self.hue,
            self.light,
            self.dark,
            self.reverse,
        )

    def __call__(self, n: int) -> Sequence[RGBHexColor]:
        return self._chmap.discrete_palette(n)


def identity_pal() -> Callable[[T], T]:
    """
    Create palette that maps values onto themselves

    Returns
    -------
    out : function
        Palette function that takes a value or sequence of values
        and returns the same values.

    Examples
    --------
    >>> palette = identity_pal()
    >>> palette(9)
    9
    >>> palette([2, 4, 6])
    [2, 4, 6]
    """
    return identity


@dataclass
class none_pal(_discrete_pal):
    """
    Discrete palette that returns only None values

    Example
    -------
    >>> palette = none_pal()
    >>> palette(5)
    [None, None, None, None, None]
    """

    def __call__(self, n: int) -> Sequence[None]:
        return [None] * n



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import TYPE_CHECKING

import dateutil.rrule as rr

if TYPE_CHECKING:
    from mizani.typing import (
        TzInfo,
    )


class DateFrequency(IntEnum):
    """
    Date Frequency

    Matching the dateutils constants
    """

    __order__ = "YEARLY MONTHLY DAILY HOURLY MINUTELY SECONDLY MICROSECONDLY"
    YEARLY = rr.YEARLY
    MONTHLY = rr.MONTHLY
    DAILY = rr.DAILY
    HOURLY = rr.HOURLY
    MINUTELY = rr.MINUTELY
    SECONDLY = rr.SECONDLY
    MICROSECONDLY = SECONDLY + 1


DF = DateFrequency


@dataclass
class date_breaks_info:
    """
    Information required to generate sequence of date breaks
    """

    frequency: DateFrequency
    n: int
    width: int
    start: datetime
    until: datetime
    tz: TzInfo | None
